import threading
import time
import curses
import os
import sys
import types
import signal
import kaa.tools
import kaa.log
from kaa import config
from kaa import options
from kaa import version
from kaa.cui import app, keydef

from kaa import fileio

CURSES_MOUSEINTERVAL = 200
CURSES_ESCDELAY = '50'


def _init(stdscr):
    if not hasattr(stdscr, 'get_wch'):
        raise RuntimeError(
            '''Wide characters are not supported by curses module.
Please install development package for ncursesw library and rebuild Python.

See https://pypi.python.org/pypi/kaaedit/#requirements for detail.
''')

    curses.start_color()
    curses.use_default_colors()

    curses.raw()
    curses.nonl()
    stdscr.leaveok(1)
#    curses.mousemask(curses.ALL_MOUSE_EVENTS)
#    curses.mouseinterval(CURSES_MOUSEINTERVAL)


def _restore():
    curses.noraw()
    curses.nl()


def run_userinit(fname):
    if os.path.isfile(fname):
        with open(fname) as f:
            src = f.read()

        code = compile(src, fname, 'exec')
        module = types.ModuleType('__kaa__')
        exec(code, module.__dict__)

        sys.modules['__kaa__'] = module

URL_CHECKVERSION_HOST = 'http://www.gembook.org'
URL_CHECKVERSION_PATH = '/kaa_checkversion/version_latest'

CHECK_DURARION = 60 * 60 * 24  # check once a day.


def _download_version_no():
    import urllib.request
    import re
    try:
        url = URL_CHECKVERSION_HOST + URL_CHECKVERSION_PATH
        f = urllib.request.urlopen(url, timeout=60)
        s = str(f.read(), 'ascii', errors='replace')
    except Exception:
        return

    m = re.match(r'(\d+)\.(\d+)\.(\d+)', s)
    if m:
        latest = tuple(int(v) for v in m.groups())
        if latest > version.KAA_VERSION:
            kaa.app.messagebar.set_message('')
            kaa.app.call_later(1, kaa.app.messagebar.set_message,
                               'New version of kaa has released.')


def _check_newversion():
    last_checked = kaa.app.config.load_value('time_check_version', 0)
    now = time.time()
    if now - last_checked < CHECK_DURARION:
        return

    kaa.app.config.save_value('time_check_version', now)
    # todo: should not rum in thread
    threading.Thread(target=_download_version_no).start()

opt = None


def main(stdscr):
    conf = config.Config(opt)
    sys.path.insert(0, conf.KAADIR)

    kaa.log.init(conf.LOGDIR)

    _init(stdscr)
    try:
        keydef.init(conf)

        kaa.app = app.CuiApp(conf)
        kaa.app.storage = fileio.FileStorage()

        if not opt.no_init:
            fname = opt.init_script
            if not fname:
                fname = os.path.join(kaa.app.config.KAADIR, '__kaa__.py')
            run_userinit(fname)

        from kaa.cui import frame
        mainframe = frame.MainFrame(stdscr)
        kaa.app.init(mainframe)

        if not opt.file:
            # no file args. show new document.
            doc = kaa.app.storage.newfile(temporary=True)
            kaa.app.show_doc(doc)
        else:
            dirname = None
            for filename in opt.file:
                if os.path.isdir(filename):
                    if not dirname:
                        dirname = filename
                else:
                    doc = kaa.app.storage.openfile(filename)
                    editor = kaa.app.show_doc(doc)
                    kaa.app.file_commands.restore_file_loc(editor)

            if dirname:
                from kaa.ui.selectfile import selectfile

                def cb(filename, encoding, newline):
                    if filename:
                        doc = kaa.app.storage.openfile(
                            filename, encoding, newline)
                        kaa.app.show_doc(doc)
                    else:
                        if not kaa.app.mainframe.childframes:
                            doc = kaa.app.storage.newfile(temporary=True)
                            kaa.app.show_doc(doc)

                selectfile.show_fileopen(dirname, cb)

        if opt.command:
            cmd = kaa.app.focus.document.mode.get_command(
                opt.command)
            is_available, command = cmd
            if command:
                command(kaa.app.focus)
            else:
                sys.exit('Unknown command: {}'.format(opt.command))

#        kaa.app.call_later(10, _check_newversion)
        kaa.app.run()
        mainframe.destroy()
        kaa.app.on_shutdown()

    finally:
        _restore()


def handle_term(signum, frame):
    sys.exit(signum + 0x80)

COLOR_ENVS = ('COLORTERM', 'XTERM_VERSION', 'ROXTERM_ID',
              'KONSOLE_DBUS_SESSION')


def _init_term():
    if opt.term:
        os.environ['TERM'] = opt.term

    # http://fedoraproject.org/wiki/Features/256_Color_Terminals
    for env in COLOR_ENVS:
        if os.environ.get(env):
            has_color = True
            break
    else:
        has_color = False

    if has_color:
        term = os.environ.get('TERM')

        if term in ('xterm', 'screen', 'Eterm'):
            term = term + '-256color'
            os.environ['TERM'] = term

        if os.environ.get('TERM', '') == 'screen-256color':
            termcap = os.environ.get('TERMCAP')
            if termcap:
                os.environ['TERMCAP'] = termcap.replace('Co#8', 'Co#256')


def run():
    if sys.version_info[:2] < (3, 3):
        raise RuntimeError('kaa requires Python 3.3 or later')

    if not getattr(__builtins__, 'kaa_freeze', False):
        try:
            setproctitle = __import__('setproctitle')
            setproctitle.setproctitle('kaa')
        except:
            pass

    signal.signal(signal.SIGTERM, handle_term)

    parser = options.build_parser()

    global opt
    opt = parser.parse_args()

    if opt.show_version:
        print(version.version_info())
        return

    if not os.environ.get('ESCDELAY'):
        os.environ['ESCDELAY'] = CURSES_ESCDELAY

    _init_term()
    curses.wrapper(main)


if __name__ == '__main__':
    run()



from unittest.mock import patch

import kaa
from kaa.ui.dialog import dialogmode
from kaa.theme import Theme, Style

import kaa_testutils


class TestDialogCursor(kaa_testutils._TestScreenBase):

    def _getcursor(self, s):
        wnd = self._getwnd(s)
        return dialogmode.DialogCursor(wnd, [dialogmode.MarkRange('mark1'), dialogmode.MarkRange('mark2')])

    def test_cursorrange(self):
        cursor = self._getcursor("0123456789")
        cursor.wnd.document.marks['mark1'] = (1, 3)
        cursor.wnd.document.marks['mark2'] = (5, 10)

        cursor.setpos(1)
        cursor.left()
        assert cursor.pos == 1
        cursor.right()
        assert cursor.pos == 2
        cursor.right()
        assert cursor.pos == 3
        cursor.right()
        assert cursor.pos == 3

        cursor.setpos(5)
        cursor.left()
        assert cursor.pos == 5
        cursor.setpos(10)
        cursor.right()
        assert cursor.pos == 10


class TestFormBuilder(kaa_testutils._TestScreenBase):

    def _getmodeclass(self):
        return dialogmode.DialogMode

    def test_build(self):
        doc = self._getdoc('')

        f = dialogmode.FormBuilder(doc)
        f.append_text('default', 'abc')

        assert doc.gettext(0, doc.endpos()) == 'abc'
        assert doc.styles.getints(0, doc.endpos()) == [0] * doc.endpos()

    def test_build_mark(self):
        doc = self._getdoc('')

        f = dialogmode.FormBuilder(doc)
        f.append_text('default', 'abc', 'mark1', 'mark2')
        f.append_text('default', 'def', 'mark3', 'mark4')

        assert doc.gettext(doc.marks['mark3'], doc.marks['mark4']) == 'def'

    def test_build_shortcut(self):
        kaa.app.DEFAULT_THEME = 'basic'

        doc = self._getdoc('')

        doc.mode.themes = [{'basic': [
            Style('default', 'default', 'default', False, False),
            Style('shortcut', 'default', 'default', False, False),
        ]}]
        doc.mode._build_theme()

        f = dialogmode.FormBuilder(doc)

        def cb(wnd):
            pass
        f.append_text('default', 'ab&cdef', on_shortcut=cb,
                      shortcut_style='shortcut', shortcut_mark='mark1')

        assert doc.gettext(0, doc.endpos()) == 'abcdef'

        id1 = doc.mode.get_styleid('default')
        id2 = doc.mode.get_styleid('shortcut')
        assert doc.styles.getints(
            0, doc.endpos()) == [id1, id1, id2, id1, id1, id1]
        assert doc.marks['mark1'] == 2



from unittest.mock import patch
import kaa_testutils
from kaa.filetype.python import pythonmode


class TestPythonMode(kaa_testutils._TestScreenBase):
    DEFAULTMODECLASS = pythonmode.PythonMode

    def test_get_headers(self):
        script = '''
class spam:
    def ham(self, \, x=(def xxx)):
        'def'
# class xyz
def egg():
   """\\\\"""
def bacon():
'''
        doc = self._getdoc(script)
        tokens = [t for t in doc.mode.get_headers()]
        assert tokens == [
            ('namespace', None, 'spam', 'spam', None, 1),
            ('function', tokens[0], 'spam.ham', 'ham()', None, 13),
            ('function', None, 'egg', 'egg()', None, 74),
            ('function', None, 'bacon', 'bacon()', None, 97)]

    def test_pyindent(self):
        # test block begins
        script = '''
abc:
'''
        wnd = self._getwnd(script)
        with patch.object(wnd.cursor, 'pos', new=5):
            wnd.document.mode.on_auto_indent(wnd)
        assert wnd.document.gettext(0, wnd.document.endpos()
                                    ) == '\nabc:\n    \n'

        # test un-closed parenthesis
        script = '''
abc(
'''
        wnd = self._getwnd(script)
        with patch.object(wnd.cursor, 'pos', new=5):
            wnd.document.mode.on_auto_indent(wnd)
        assert wnd.document.gettext(0, wnd.document.endpos()
                                    ) == '\nabc(\n    \n'

        # test balanced parenthesis
        script = '''
abc()
'''
        wnd = self._getwnd(script)
        with patch.object(wnd.cursor, 'pos', new=6):
            wnd.document.mode.on_auto_indent(wnd)
        assert wnd.document.gettext(0, wnd.document.endpos()
                                    ) == '\nabc()\n\n'

        # test closed parenthesis
        script = '''
  abc
    abc)
'''

        wnd = self._getwnd(script)
        with patch.object(wnd.cursor, 'pos', new=15):
            wnd.document.mode.on_auto_indent(wnd)
        assert wnd.document.gettext(0, wnd.document.endpos()
                                    ) == '\n  abc\n    abc)\n    \n'

        # test closed parenthesis, but no dedent required
        script = '''
  abc
  )
'''
        wnd = self._getwnd(script)
        with patch.object(wnd.cursor, 'pos', new=10):
            wnd.document.mode.on_auto_indent(wnd)
        assert wnd.document.gettext(0, wnd.document.endpos()
                                    ) == '\n  abc\n  )\n  \n'

        # test comment
        script = '''
abc#(
'''
        wnd = self._getwnd(script)
        with patch.object(wnd.cursor, 'pos', new=6):
            wnd.document.mode.on_auto_indent(wnd)
        assert wnd.document.gettext(0, wnd.document.endpos()
                                    ) == '\nabc#(\n\n'

        # test string
        script = '''
abc"(
'''
        wnd = self._getwnd(script)
        with patch.object(wnd.cursor, 'pos', new=6):
            wnd.document.mode.on_auto_indent(wnd)
        assert wnd.document.gettext(0, wnd.document.endpos()
                                    ) == '\nabc"(\n\n'

        # test string closed
        script = '''
abc"""-"""(
'''
        wnd = self._getwnd(script)
        with patch.object(wnd.cursor, 'pos', new=12):
            wnd.document.mode.on_auto_indent(wnd)
        assert wnd.document.gettext(0, wnd.document.endpos()
                                    ) == '\nabc"""-"""(\n    \n'



import os
import kaa
import kaa.log
import kaa.utils
from kaa.command import Commands
from kaa.command import commandid
from kaa.command import norec
from kaa.command import norerun


class FileCommands(Commands):

    def _show_msgbox(self, caption, options, callback,
                     keys=None, border=False):

        from kaa.ui.msgbox import msgboxmode
        return msgboxmode.MsgBoxMode.show_msgbox(
            caption, options, callback, keys, border)

    @commandid('file.new')
    @norec
    @norerun
    def file_new(self, wnd):
        doc = kaa.app.storage.newfile()
        kaa.app.show_doc(doc)

    def _find_file_doc(self, filename):
        for frame in kaa.app.get_frames():
            for doc in frame.get_documents():
                if doc.get_filename() == filename:
                    MAX_FILENAMEMSG_LEN = 50
                    if len(filename) > MAX_FILENAMEMSG_LEN:
                        filename = '...{}'.format(
                            filename[MAX_FILENAMEMSG_LEN * -1:])

                    kaa.app.messagebar.set_message(
                        "`{}` is already opened.".format(filename))
                    return doc

    def _activate_file(self, filename):
        doc = self._find_file_doc(filename)
        if doc:
            if doc.wnds:
                wnd = doc.wnds[0]
                wnd.get_label('frame').bring_top()
                wnd.activate()
            return True

    def restore_file_loc(self, wnd):
        if wnd.document.fileinfo:
            disp = kaa.app.config.hist('filedisp').find(
                wnd.document.fileinfo.fullpathname)
            if disp:
                loc = disp.get('pos', 0)
                loc = min(loc, wnd.document.endpos())
                if wnd.document.geteol(loc) != wnd.document.endpos():
                    wnd.cursor.setpos(loc, top=True)
                else:
                    wnd.cursor.setpos(loc, middle=True)

    @commandid('file.open')
    @norec
    @norerun
    def file_open(self, wnd, filename=''):
        def cb(filename, encoding, newline):
            if not filename:
                return

            if self._activate_file(filename):
                return

            doc = kaa.app.storage.openfile(filename, encoding, newline)
            editor = kaa.app.show_doc(doc)
            self.restore_file_loc(editor)

        if not filename:
            if wnd and wnd.document.fileinfo:
                filename = wnd.document.fileinfo.dirname

        from kaa.ui.selectfile import selectfile
        selectfile.show_fileopen(filename, cb)

    @commandid('file.info')
    @norec
    @norerun
    def file_info(self, wnd):
        from kaa.ui.fileinfo import fileinfomode
        fileinfomode.show_fileinfo(wnd)

    @commandid('file.viewdiff')
    @norec
    @norerun
    def file_viewdiff(self, wnd, callback=None):
        if wnd.document.fileinfo and wnd.document.fileinfo.fullpathname:
            from kaa.ui.viewdiff import viewdiffmode
            viewdiffmode.view_doc_diff(wnd.document, callback=None)

    @commandid('file.save')
    @norec
    @norerun
    def file_save(self, wnd, filename=None, saved=None, document=None,
                  encoding=None, newline=None):
        "Save document"
        try:
            if not document:
                document = wnd.document
            if not filename:
                filename = document.get_filename()

            if filename:
                kaa.app.storage.save_document(
                    document, filename, encoding, newline)
                # notify file saved
                if saved:
                    saved(canceled=False)
            else:
                # no file name. Show save_as dialog.
                self.file_saveas(wnd, saved=saved, document=document)

        except Exception as e:
            kaa.log.exception('File write error:')
            self._show_msgbox(
                'File write error: ' + str(e), ['&Ok'],
                lambda c: saved(canceled=True) if saved else None,
                keys=['\r', '\n'])

    @commandid('file.saveas')
    @norec
    @norerun
    def file_saveas(self, wnd, saved=None, document=None):
        "Show save_as dialog and save to the specified file."
        def cb(filename, enc, newline):
            if filename:
                self.file_save(wnd, filename, saved=saved, document=document,
                               encoding=enc, newline=newline)
            else:
                if saved:
                    saved(canceled=True)

        if not document:
            document = wnd.document

        from kaa.ui.selectfile import selectfile

        if document.fileinfo:
            newline = document.fileinfo.newline
            encoding = document.fileinfo.encoding
        else:
            newline = encoding = None

        selectfile.show_filesaveas(
            document.get_filename(), newline, encoding, cb)

    def ask_doc_close(self, wnd, document, callback, msg):
        def choice(c):
            if c == 'y':
                self.file_save(wnd, saved=callback, document=document)
            elif c == 'n':
                callback(canceled=False)
            elif c == 'd':
                def cb():
                    return self.ask_doc_close(
                        wnd, document, callback, msg)

                if document.fileinfo and document.fileinfo.fullpathname:
                    from kaa.ui.viewdiff import viewdiffmode
                    viewdiffmode.view_doc_diff(document, callback=cb)
                else:
                    cb()

        if (document.mode.DOCUMENT_MODE
                and document.undo
                and document.undo.is_dirty()):

            items = ['&Yes', '&No', '&Cancel']
            if document.fileinfo and document.fileinfo.fullpathname:
                items.append('View &Diff')

            self._show_msgbox(
                '{} [{}]: '.format(
                    msg, document.get_title()),
                items, choice)
        else:
            callback(canceled=False)

    def save_documents(self, wnd, docs, callback,
                       msg='Save file before close?', force=False):

        docs = list(docs)
        activeframe = kaa.app.get_activeframe()

        def _save_documents(canceled):
            if canceled:
                if activeframe and not activeframe.closed:
                    kaa.app.set_focus(activeframe)
                callback(canceled=True)
                return

            if not docs:
                if activeframe and not activeframe.closed:
                    kaa.app.set_focus(activeframe)

                if callback:
                    callback(canceled=False)
            else:
                doc = docs.pop()
                if doc.wnds:
                    w = doc.wnds[0]
                    if w and not w.closed:
                        w.activate()

                if force:
                    self.file_save(wnd, saved=_save_documents, document=doc)
                else:
                    self.ask_doc_close(wnd, doc, _save_documents, msg)

        _save_documents(canceled=False)

    def get_closed_docs(self, editors):
        alldocs = {e.document for e in editors}
        docs = []

        for doc in alldocs:
            doc_editors = set(doc.wnds) - editors
            if not doc_editors:
                docs.append(doc)
        return docs

    def close_frame(self, frame, callback):

        editors = {e for e in frame.get_editors()}
        docs = self.get_closed_docs(editors)

        def cb(canceled):
            if not canceled:
                frame.destroy()
                callback(canceled=False)

        self.save_documents(frame, docs, cb, 'Save file before close?')

    @commandid('file.close')
    @norec
    @norerun
    def file_close(self, wnd, callback=None):
        "Close active frame"

        import kaa.fileio

        frame = wnd.get_label('frame')

        def saved(canceled):
            if canceled:
                return

            if frame.mainframe.childframes:
                f = frame.mainframe.childframes[0]
                f.bring_top()
                kaa.app.set_focus(f)
            else:
                doc = kaa.app.storage.newfile(temporary=True)
                kaa.app.show_doc(doc)

        self.close_frame(frame, saved)

    def get_current_documents(self, wnd):
        docs = set()
        for frame in wnd.mainframe.childframes:
            editors = {e for e in frame.get_editors()}
            docs |= {
                e.document for e in editors if e.document.mode.DOCUMENT_MODE}
        return docs

    @commandid('file.save.all')
    @norec
    @norerun
    def file_saveall(self, wnd):

        def saved(canceled):
            pass

        docs = self.get_current_documents(wnd)
        docs = [doc for doc in docs if doc.undo and doc.undo.is_dirty()]
        if docs:
            self.save_documents(wnd, docs, saved, force=True)

    @commandid('file.close.all')
    @norec
    @norerun
    def file_closeall(self, wnd):

        def callback(canceled):
            if canceled:
                return

            import kaa.fileio
            if frames:
                f = frames.pop()
                self.close_frame(f, callback)
            else:
                doc = kaa.app.storage.newfile(temporary=True)
                kaa.app.show_doc(doc)

        frames = wnd.mainframe.childframes[:]
        callback(False)

    def _select_recentry_used_files(self, callback):
        files = []
        for p, info in kaa.app.config.hist('filename').get():
            path = kaa.utils.shorten_filename(p)
            files.append(path if len(path) < len(p) else p)

        from kaa.ui.selectlist import filterlist
        filterlist.show_listdlg('Recently used files:',
                                files, callback)

    @commandid('file.recently-used-files')
    @norec
    @norerun
    def file_recently_used_files(self, wnd):
        def cb(filename):
            if filename:
                filename = os.path.abspath(filename)
                if self._activate_file(filename):
                    return

                self.file_open(wnd, filename)

        self._select_recentry_used_files(cb)

    def _select_recentry_used_dirs(self, callback):
        files = []
        for p, info in kaa.app.config.hist('dirname').get():
            path = kaa.utils.shorten_filename(p)
            files.append(path if len(path) < len(p) else p)

        from kaa.ui.selectlist import filterlist
        filterlist.show_listdlg('Recently used directories:',
                                files, callback)

    @commandid('file.recently-used-directories')
    @norec
    @norerun
    def file_recently_used_dirs(self, wnd):
        def cb(filename):
            if filename:
                self.file_open(wnd, filename)

        self._select_recentry_used_dirs(cb)

    @commandid('file.quit')
    @norec
    @norerun
    def file_quit(self, wnd):
        def saved(canceled):
            if not canceled:
                try:
                    for frame in wnd.mainframe.childframes:
                        frame.destroy()
                except Exception:
                    kaa.log.exception('Error in file.quit:')
                kaa.app.quit()

        docs = self.get_current_documents(wnd)
        self.save_documents(wnd, docs, saved, 'Save file before close?')

    def reload_file(self, document):
        fileinfo = document.fileinfo
        newdoc = kaa.app.storage.openfile(
            fileinfo.fullpathname, fileinfo.encoding, fileinfo.newline,
            filemustexists=True)

        for w in document.wnds[:]:
            w.show_doc(newdoc)

        if not document.closed:
            document.close()

    def notify_fileupdated(self, document):
        def choice(c):
            if c == 'y':
                self.reload_file(document)
            elif c == 'n':
                return
            elif c == 'd':
                def cb():
                    return self.notify_fileupdated(document)

                if document.fileinfo and document.fileinfo.fullpathname:
                    from kaa.ui.viewdiff import viewdiffmode
                    viewdiffmode.view_doc_diff(
                        document, callback=cb)
                else:
                    cb()
            else:
                self.notify_fileupdated(document)

        items = ['&Yes', '&No', 'View &Diff']
        msg = 'File [{}] has been updated by other process. Reload file?: '
        self._show_msgbox(msg.format(document.get_title()),
                          items, choice)

    def can_close_wnd(self, wnd, cb):
        if len(wnd.document.wnds) == 1:
            self.ask_doc_close(
                wnd, wnd.document, cb, 'Save file before close?')
        else:
            cb(False)

    def _open_file_to_wnd(self, wnd, func):
        def cb(filename, encoding, newline):
            if not filename:
                return

            doc = self._find_file_doc(filename)
            if not doc:
                doc = kaa.app.storage.openfile(filename, encoding, newline)

            wnd.show_doc(doc)
            self.restore_file_loc(wnd)

        def saved(canceled):
            if canceled:
                return
            func(cb)

        self.can_close_wnd(wnd, saved)

    @commandid('file.new-to')
    @norec
    @norerun
    def file_new_to(self, wnd):
        def openfile(cb):
            doc = kaa.app.storage.newfile()
            wnd.show_doc(doc)

        self._open_file_to_wnd(wnd, openfile)

    @commandid('file.open-to')
    @norec
    @norerun
    def file_open_to(self, wnd):
        def openfile(cb):
            if wnd.document.fileinfo:
                filename = wnd.document.fileinfo.dirname
            else:
                filename = ''
            from kaa.ui.selectfile import selectfile
            selectfile.show_fileopen(filename, cb)

        self._open_file_to_wnd(wnd, openfile)

    def _show_selected_recentlyfile(self, cb, filename):
        if filename:
            filename = os.path.abspath(filename)

            from kaa.ui.selectfile import selectfile
            selectfile.show_fileopen(filename, cb)

    @commandid('file.recently-used-files-to')
    @norec
    @norerun
    def file_recently_used_files_to(self, wnd):
        def selectfile(cb):
            self._select_recentry_used_files(
                lambda filename: self._show_selected_recentlyfile(cb, filename))

        self._open_file_to_wnd(wnd, selectfile)

    @commandid('file.recently-used-directories-to')
    @norec
    @norerun
    def file_recently_used_dirs_to(self, wnd):
        def selectdir(cb):
            self._select_recentry_used_dirs(
                lambda filename: self._show_selected_recentlyfile(cb, filename))

        self._open_file_to_wnd(wnd, selectdir)



import kaa
import os
import functools


def ignore_errors(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except Exception:
            try:
                kaa.log.exception('')
            except Exception:
                # Ignore error completely...
                pass

    return wrapper


def split_existing_dirs(path):
    path, p = os.path.split(path)
    rights = [p]
    while True:
        if not path or os.path.isdir(path):
            if rights:
                rest = os.path.join(*rights)
            else:
                rest = ''
            return path, rest
        path, right = os.path.split(path)
        if right:
            rights.insert(0, right)


def shorten_filename(f):
    rel = os.path.relpath(f)
    home = os.path.join('~', os.path.relpath(f, os.path.expanduser('~')))

    tp = sorted(((n.count(os.sep), len(n), n) for n in (f, rel, home)))
    return tp[0][2]



import os
import curses
import subprocess
import signal
import time
import kaa
from kaa import document
from kaa.command import Commands
from kaa.command import commandid
from kaa.command import norec
from kaa.command import norerun


class ToolCommands(Commands):

    @commandid('python.console')
    @norerun
    def pythonconsole(self, wnd):
        from kaa.ui.pyconsole import pythonconsolemode
        pythonconsolemode.show_console()

    @commandid('python.debugger.run')
    @norerun
    def pythonchilddebugger(self, wnd):
        from kaa.ui.pythondebug import port
        port.run()

    @commandid('python.debugger.server')
    @norerun
    def pythondebuggerserver(self, wnd):
        from kaa.ui.pythondebug import port
        port.init_server()

    @commandid('command.rerun')
    @norerun
    def reruncommand(self, wnd):
        mode = wnd.document.mode
        if not kaa.app.lastcommands:
            return
        n_repeat, commandids = kaa.app.lastcommands
        wnd.set_command_repeat(n_repeat)
        try:
            for cmdid in commandids:
                is_available, command = mode.get_command(cmdid)
                if not command:
                    msg = 'command {!r} is not registered.'.format(cmdid)
                    kaa.app.messagebar.set_message(msg)
                    kaa.log.error(msg)
                    return
                command(wnd)
        finally:
            wnd.set_command_repeat(1)

    @commandid('edit.paste-lines')
    @norerun
    def edit_pastelines(self, wnd):
        from kaa.ui.pastelines import pastelinesmode
        doc = pastelinesmode.PasteLinesMode.build(wnd)

        kaa.app.show_dialog(doc)

    @commandid('tools.execute-shell-command')
    @norerun
    def execute_shell_command(self, wnd):
        def callback(w, s):
            s = s.strip()
            if s:
                kaa.app.config.hist('shellcommand').add(s)
                ret = subprocess.check_output(
                    s, stderr=subprocess.STDOUT,
                    shell=True,
                    universal_newlines=True)

                wnd.document.mode.put_string(wnd, ret)
                wnd.screen.selection.clear()

                kaa.app.messagebar.set_message(
                    "{} letters inserted".format(len(ret)))

        hist = [s for s, info in kaa.app.config.hist('shellcommand').get()]
        from kaa.ui.inputline import inputlinemode
        doc = inputlinemode.InputlineMode.build('Shell command:',
                                                callback, history=hist)
        kaa.app.messagebar.set_message('Execute shell command')

        kaa.app.show_dialog(doc)

    def _exec_cmd(self, cmd):
        # todo: move to util
        import select
        import errno
        # http://stackoverflow.com/a/12207447
        # This page sujests bare os pipes is never buffered.
        # Is it true? Looks works fine, though.
        master, slave = os.pipe()
        # todo: Make these handles not inherited in Python 3.3.
        with subprocess.Popen(cmd, shell=True, stdout=slave,
                              stderr=slave, bufsize=1,
                              universal_newlines=True) as process:

            os.close(slave)

            output = os.fdopen(master)
            try:
                lines = []
                while process.poll() is None:
                    rlist, _, _ = select.select([master], [], [])
                    for f in rlist:
                        l = output.readline()
                        lines.append(l)
                        print(l, end='')
            finally:
                output.close()

            retcode = process.poll()
            if retcode:
                code = errno.errorcode.get(retcode)
                if code:
                    code = '(%s):' % code
                else:
                    code = ''

                msg = os.strerror(retcode)
                kaa.app.messagebar.set_message(
                    '{code}{msg}'.format(
                        code=code, msg=msg, retcode=retcode))

            return ''.join(lines)

    @commandid('tools.make')
    @norerun
    def execute_make(self, wnd):
        def callback(w, s):
            s = s.strip()
            if s:
                kaa.app.config.hist('makecommand').add(s)
                # todo: move these lines to kaa.cui.*
                curses.def_prog_mode()
                curses.endwin()
                try:
                    ret = self._exec_cmd(s)
                finally:
                    curses.reset_prog_mode()
                    wnd.mainframe.refresh()

                from kaa.ui.makeoutput import makeoutputmode
                makeoutputmode.show(s, ret)

        from kaa.ui.inputline import inputlinemode
        hist = [s for s, info in kaa.app.config.hist('makecommand').get()]
        value = 'make' if not hist else hist[0]
        doc = inputlinemode.InputlineMode.build('Make command:', callback,
                                                history=hist, value=value)
        kaa.app.messagebar.set_message('Execute command')

        kaa.app.show_dialog(doc)

    @commandid('tools.spellchecker')
    @norec
    @norerun
    def spellchecker(self, wnd):
        try:
            pass
        except ImportError:
            kaa.app.messagebar.set_message(
                'PyEnchant module is not installed.')
            return

        from kaa.ui.spellchecker import spellcheckermode
        spellcheckermode.run_spellchecker(wnd)

    @commandid('tools.grep')
    @norec
    @norerun
    def showgrep(self, wnd):
        from kaa.ui.grep import grepdlgmode

        doc = document.Document()
        mode = grepdlgmode.GrepDlgMode(wnd)
        doc.setmode(mode)
        mode.build_document()

        kaa.app.show_dialog(doc)

    @commandid('tools.suspend')
    @norec
    @norerun
    def suspend(self, wnd):
        with kaa.app.restore_teminal():
            os.kill(os.getpid(), signal.SIGSTOP)
            time.sleep(0.01)  # sleep to ensure signal derivered



import os
import signal
import kaa
from kaa import document
from kaa.ui.dialog import dialogmode
from kaa.theme import Overlay
from kaa.theme import Style
from kaa.command import commandid, norec, norerun
from kaa.keyboard import *
from kaa.ui.selectlist import selectlist
import kaa.ui.pythondebug.port
from kaa.ui.inputline import inputlinemode


def show_expr(port, depth):
    def callback(w, s):
        s = s.strip()
        if s:
            kaa.app.config.hist('pythondebug_expr').add(s)
            port.show_expr(depth, s)

    hist = [s for s, info in kaa.app.config.hist('pythondebug_expr').get()]
    doc = inputlinemode.InputlineMode.build('Python expression:',
                                            callback, history=hist)
    kaa.app.show_dialog(doc)


breakpoints_keys = {
    'd': 'breakpoints.delete',
    (ctrl, 'h'): 'breakpoints.delete',
    delete: 'breakpoints.delete',
    backspace: 'breakpoints.delete',
}


class BreakPoints(selectlist.SelectItemList):
    SEP = '\n'
    caption = 'Break points'
    border = True

    @classmethod
    def build(cls, port):
        doc = super().build()
        doc.mode.show_breakpoints(port)
        kaa.app.messagebar.set_message("Hit 'd' key to remove breakpoint")
        return doc

    def init_keybind(self):
        super().init_keybind()
        self.keybind.add_keybind(breakpoints_keys)

    def show_breakpoints(self, port):
        self.port = port
        breakpoints = self.port.get_breakpoints()
        breakpoints.sort(key=lambda o: (o.filename, o.lineno))
        items = []
        for bp in breakpoints:
            caption = '{bp.filename}:{bp.lineno}'.format(bp=bp)
            c = selectlist.SelectItem(
                'selectitem', 'selectitem-active', caption, bp)
            items.append(c)

        self.update_doc(items)

    def on_esc_pressed(self, wnd, event):
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message('')

    @commandid('breakpoints.delete')
    @norec
    @norerun
    def delete_bp(self, wnd):
        if not self.cursel:
            return
        bp = self.cursel.value

        index = self.items.index(self.cursel)
        items = [item for item in self.items if item is not self.cursel]
        self.update_doc(items)
        if not self.items:
            sel = None
        elif index < len(self.items):
            sel = self.items[index]
        else:
            sel = self.items[-1]

        self.update_sel(wnd, sel)
        self.port.del_breakpoint(bp)
        self.port.display_breakpoints()

DebugThemes = {
    'basic': [
        Style('line', 'Orange', None),
        Style('filename', 'Cyan', None),
        Style('lineno', 'Cyan', None),
        Style('funcname', 'Cyan', None),
        Style('dirname', 'Cyan', None),
        Style('status', 'Base3', 'Red', nowrap=True),
        Overlay('current_stack', 'Black', 'Yellow'),
    ],
}


callstack_keys = {
    down: 'callstack_keys.next',
    (ctrl, 'n'): 'callstack_keys.next',
    (ctrl, 'f'): 'callstack_keys.next',
    tab: 'callstack_keys.next',
    up: 'callstack_keys.prev',
    (ctrl, 'p'): 'callstack_keys.prev',
    (ctrl, 'b'): 'callstack_keys.prev',
    (shift, tab): 'callstack_keys.prev',
}


class PythonStackList(dialogmode.DialogMode):
    SHOW_CURSOR = False
    autoshrink = True
    cursel = 0

    def init_themes(self):
        super().init_themes()
        self.themes.append(DebugThemes)

    def calc_height(self, wnd):
        height = wnd.screen.get_total_height(wnd.mainframe.height // 2)
        return height

    def build(self, stack):
        self.document.marks.clear()
        self.document.delete(0, self.document.endpos())

        self.stack = tuple(stack or ())
        with dialogmode.FormBuilder(self.document) as f:
            for n, (fname, lno, funcname, lines) in enumerate(self.stack):
                s = self.document.endpos()

                f.append_text('funcname', funcname.replace('&', '&&'))
                f.append_text('default', ':')

                dirname, filename = os.path.split(fname)
                f.append_text('filename', filename.replace('&', '&&'))
                f.append_text('default', ':')

                f.append_text('lineno', str(lno))
                f.append_text('default', ':')

                f.append_text('dirname', dirname.replace('&', '&&') + '\n')

                line = (lines[0] if lines else '').strip()
                if not line:
                    line = '(empty line)'
                f.append_text('line', line.replace('&', '&&') + '\n')

                t = self.document.endpos()
                self.document.marks[('stack', n)] = (s, t)

    def update_sel(self, wnd, n):
        if self.cursel != n:
            f, t = self.document.marks.get(
                ('stack', self.cursel), (None, None))
            if f is not None:
                wnd.set_line_overlay(f, None)

        f, t = self.document.marks.get(('stack', n), (None, None))
        if f is not None:
            wnd.set_line_overlay(f, 'current_stack')

        self.cursel = n
        if n < len(self.stack):
            (fname, lno, funcname, lines) = self.stack[n]

            if f is not None:
                wnd.cursor.setpos(f, bottom=True)

            wnd.activate()
            debugpanel = wnd.get_label('editor')
            debugpanel.document.mode.show_line(fname, lno)
            debugpanel.activate()

    @commandid('callstack_keys.prev')
    @norec
    @norerun
    def stack_prev(self, wnd):
        if self.cursel:
            self.update_sel(wnd, self.cursel - 1)

    @commandid('callstack_keys.next')
    @norec
    @norerun
    def stack_next(self, wnd):
        if self.cursel < len(self.stack) - 1:
            self.update_sel(wnd, self.cursel + 1)


class PythonDebuggerPanel(dialogmode.DialogMode):
    SHOW_CURSOR = False
    autoshrink = True

    def init_themes(self):
        super().init_themes()
        self.themes.append(DebugThemes)

    def init_keybind(self):
        super().init_keybind()
        self.keybind.add_keybind(callstack_keys)

    def setup(self):
        super().setup()
        self.updated_wnds = set()

    def close(self):
        self._restore_highlight()
        super().close()
        self.port = None

    def on_str(self, wnd, s, overwrite=False):
        pass

    def build(self, stack):
        self.document.marks.clear()
        self.document.delete(0, self.document.endpos())

        self.stack = tuple(stack or ())
        with dialogmode.FormBuilder(self.document) as f:
            f.append_text('button', '[&Step]',
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_step)

            f.append_text('button', '[&Next]',
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_next)

            f.append_text('button', '[&Return]',
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_return)

            f.append_text('button', '[&Continue]',
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_continue)

            f.append_text('button', '[&Expr]',
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_expr)

            f.append_text('button', '[&Kill]',
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_kill)

            f.append_text('button', '[&Breakpoint]',
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_breakpoint)

            f.append_text('button', '[&Quit]',
                          shortcut_style='button.shortcut',
                          on_shortcut=self.on_quit)

            f.append_text('status', '-Waiting-', mark_pair='status')

    def update(self, wnd, stack):
        self.build(stack)
        self.show_curline(wnd)

        w = self._get_stacklist(wnd)
        w.document.mode.build(stack)
        w.document.mode.update_sel(w, 0)

    def show_curline(self, wnd):
        if self.stack:
            (fname, lno, funcname, lines) = self.stack[0]
            self.show_line(fname, lno)
            wnd.activate()

            wnd.mainframe.on_console_resized()

    def on_esc_pressed(self, wnd, event):
        popup = wnd.get_label('popup')
        popup.destroy()
        kaa.app.messagebar.set_message('')

    def set_status(self, wnd, status):
        f, t = self.document.marks['status']
        styleid = self.get_styleid('status')
        self.document.replace(f, t, status, styleid)
        wnd.clear_line_overlay()
#        self.update_sel(wnd, 0)

    def _restore_highlight(self):
        for w in self.updated_wnds:
            w.clear_line_overlay()
        self.updated_wnds = set()

    def _locate_doc(self, wnd, doc, lineno):
        self._restore_highlight()

        pos = doc.get_lineno_pos(lineno)
        tol = doc.gettol(pos)

        wnd.clear_line_overlay()
        wnd.set_line_overlay(tol, 'current-row')

        wnd.screen.locate(tol, middle=True, align_always=True)
        wnd.cursor.setpos(tol, middle=True)
        wnd.activate()
        self.updated_wnds.add(wnd)

    def show_line(self, filename, lineno):
        filename = os.path.abspath(filename)
        doc = document.Document.find_filename(filename)
        if doc:
            wnd = doc.wnds[0]
            wnd.activate()
        else:
            doc = kaa.app.storage.openfile(filename, nohist=True)
            wnd = kaa.app.show_doc(doc)

        self._locate_doc(wnd, doc, lineno)

    msg_notbreaking = 'Debug target is not idle.'

    def _cmd_not_breaked(self):
        kaa.app.messagebar.set_message(self.msg_notbreaking)

    def on_step(self, wnd):
        if self.port.is_breaking():
            self.port.set_step()
        else:
            self._cmd_not_breaked()

    def on_next(self, wnd):
        if self.port.is_breaking():
            self.port.set_next()
        else:
            self._cmd_not_breaked()

    def on_return(self, wnd):
        if self.port.is_breaking():
            self.port.set_return()
        else:
            self._cmd_not_breaked()

    def on_continue(self, wnd):
        if self.port.is_breaking():
            self.port.set_continue()
        else:
            self._cmd_not_breaked()

    def on_expr(self, wnd):
        if self.port.is_breaking():
            show_expr(self.port, self._get_stacklist(wnd).document.mode.cursel)
        else:
            self._cmd_not_breaked()

    def on_kill(self, wnd):
        if not self.port.is_breaking():
            os.kill(self.port.kaadbg_pid, signal.SIGINT)
        else:
            msg = 'Debug target is idle.'
            kaa.app.messagebar.set_message(msg)

    def on_breakpoint(self, wnd):
        doc = BreakPoints.build(self.port)
        kaa.app.show_dialog(doc)

    def on_quit(self, wnd):
        self.port.close()

    def _get_stacklist(self, wnd):
        return wnd.get_label('dlg_stacklist')

    @commandid('callstack_keys.prev')
    @norec
    @norerun
    def stack_prev(self, wnd):
        w = self._get_stacklist(wnd)
        w.document.mode.stack_prev(w)

    @commandid('callstack_keys.next')
    @norec
    @norerun
    def stack_next(self, wnd):
        w = self._get_stacklist(wnd)
        w.document.mode.stack_next(w)


def show_callstack(port, stack):
    doc = document.Document()
    doc.set_title('Python call stack')
    mode = PythonDebuggerPanel()
    mode.port = port

    doc.setmode(mode)
    mode.build(stack)

    dlg = kaa.app.show_inputline(doc)
    ret = dlg.get_label('editor')

    stacklist = document.Document()
    stacklistmode = PythonStackList()
    stacklist.setmode(stacklistmode)
    stacklistmode.build(stack)

    wnd = dlg.add_doc('dlg_stacklist', 0, stacklist)
    stacklistmode.update_sel(wnd, 0)

    mode.show_curline(ret)

    return ret



from contextlib import ExitStack
from unittest.mock import patch

import kaa
import kaa_testutils


class TestSwitchSource(kaa_testutils._TestScreenBase):

    def test_switchsource(self):
        with ExitStack() as st:
            ec = st.enter_context

            wnd = self._getwnd('')
            ec(patch.object(wnd, 'show_doc', create=True))

            frame = kaa_testutils.DmyFrame()
            frame.docs = [wnd.document]

            ec(patch('kaa.app.get_frames',
                     create=True, return_value=[frame]))

            def show_dlg(doc):
                doc.mode.callback(False)
            ec(patch('kaa.app.show_dialog', new=show_dlg, create=True))

            kaa.app.app_commands.editor_switchfile(wnd)

            args, kwargs = wnd.show_doc.call_args
            assert args[0] is wnd.document


class TestJoinWindow(kaa_testutils._TestScreenBase):

    def test_joinwindow(self):
        with ExitStack() as st:
            ec = st.enter_context

            wnd = self._getwnd('')
            ec(patch.object(wnd, 'splitter', create=True))
            ec(patch.object(kaa.app, 'get_activeframe', create=True))
            ec(patch.object(wnd, 'show_doc', create=True))

            kaa.app.app_commands.editor_joinwindow(wnd)

            args, kwargs = wnd.splitter.parent.join.call_args
            assert args[0] is wnd



from kaa.command import commandid
from kaa.command import norec
from kaa.command import norerun
from kaa import document
from kaa.ui.dialog import dialogmode
from kaa.keyboard import *

ItemListThemes = {
    'basic': [],
}

itemlist_keys = {
    left: 'itemlist.prev',
    (ctrl, 'p'): 'itemlist.prev',
    (ctrl, 'b'): 'itemlist.prev',
    (shift, tab): 'itemlist.prev',
    right: 'itemlist.next',
    (ctrl, 'n'): 'itemlist.next',
    (ctrl, 'f'): 'itemlist.next',
    tab: 'itemlist.next',
    '\r': 'itemlist.selected',
    '\n': 'itemlist.selected',
}


class ItemListMode(dialogmode.DialogMode):
    border = True
    autoshrink = True
    CAPTIONSEP = '\n'
    ITEMSEP = ' '

    @classmethod
    def build(cls, caption, items, sel, callback, onchange=None):
        doc = document.Document()
        mode = cls()
        doc.setmode(mode)

        mode.items = items
        mode.cursel = sel
        mode.callback = callback
        mode.onchange = onchange

        with dialogmode.FormBuilder(doc) as f:

            if caption:
                f.append_text('caption', caption.replace('&', '&&'))
                f.append_text('default', cls.CAPTIONSEP)

            for i, item in enumerate(items):
                f.append_text('default', item.replace('&', '&&'), mark_pair=i)
                f.append_text('default', cls.ITEMSEP)

            mode._update_style(None)
        return doc

    def _update_style(self, wnd):
        cursor = None
        for i in range(len(self.items)):
            f, t = self.document.marks[i]
            if i == self.cursel:
                style = 'activemark'
                cursor = f
            else:
                style = 'nonactivemark'
                if cursor is None:
                    cursor = f

            self.document.setstyles(f, t, self.get_styleid(style))

        if wnd and cursor is not None:
            wnd.get_label('editor').cursor.setpos(cursor)

    def init_keybind(self):
        self.keybind.add_keybind(itemlist_keys)

    def init_themes(self):
        super().init_themes()
        self.themes.append(ItemListThemes)

    def is_cursor_visible(self):
        return 0   # hide cursor

    def on_esc_pressed(self, wnd, event):
        self.cursel = None
        self.selected(wnd)

    def on_str(self, wnd, s, overwrite=False):
        cursel = self.cursel
        if cursel is None:
            cursel = -1
        c = s[-1]
        c = (c.lower(), c.upper())
        cont = True
        while cont:
            if cursel < 0:
                cont = False
            for i in range(cursel + 1, len(self.items)):
                if self.items[i].startswith(c):
                    self.cursel = i
                    self._update_style(wnd)
                    return
            cursel = -1

    @commandid('itemlist.prev')
    @norec
    @norerun
    def prev(self, wnd):
        mode = wnd.document.mode
        if mode.items:
            cur = mode.cursel
            if mode.cursel is None:
                mode.cursel = 0
            elif mode.cursel > 0:
                mode.cursel -= 1

            if cur is None or cur != mode.cursel:
                if mode.onchange:
                    mode.onchange(mode.cursel)

        wnd.document.mode._update_style(wnd)

    @commandid('itemlist.next')
    @norec
    @norerun
    def next(self, wnd):
        mode = wnd.document.mode
        if mode.items:
            cur = mode.cursel
            if mode.cursel is None:
                mode.cursel = 0
            elif mode.cursel < len(mode.items) - 1:
                mode.cursel += 1

            if cur is None or cur != mode.cursel:
                if mode.onchange:
                    mode.onchange(mode.cursel)

        wnd.document.mode._update_style(wnd)

    @commandid('itemlist.selected')
    @norec
    @norerun
    def selected(self, wnd):
        callback = wnd.document.mode.callback
        cursel = wnd.document.mode.cursel

        # Destroy popup window
        popup = wnd.get_label('popup')
        if popup:
            popup.destroy()

        callback(cursel)





class Theme:

    def __init__(self, styles):
        self.styles = {}
        self.overlays = {}
        self.add_styles(styles)

    def get_style(self, name):
        return self.styles[name]

    def add_styles(self, styles):
        for style in styles:
            if not isinstance(style, Overlay):
                # Style instance
                self.styles[style.name] = style
            else:
                self.overlays[style.name] = style

    def update(self, rhs):
        default = self.styles.get('default', None)
        styles = []
        for s in rhs:
            s = s.copy()
            if s.name == 'default':
                s.set_default_attr(default)
            styles.append(s)
        self.add_styles(styles)

    def finish_update(self):
        default = self.styles.get('default', None)
        for style in self.styles.values():
            style.set_default_attr(default)


class Style:

    def __init__(self, name, fgcolor, bgcolor, underline=False,
                 bold=False, nowrap=False, rjust=False, fillrow=False):
        self.name = name
        self.fgcolor = fgcolor
        self.bgcolor = bgcolor
        self.underline = underline
        self.bold = bold
        self.nowrap = nowrap
        self.rjust = rjust
        self.fillrow = fillrow

    def copy(self):
        # ???
        return self.__class__(
            self.name,
            self.fgcolor,
            self.bgcolor,
            self.underline,
            self.bold,
            self.nowrap,
            self.rjust,
            self.fillrow,
        )

    def set_default_attr(self, default):
        # ???
        if self.fgcolor is None:
            self.fgcolor = default.fgcolor
        if self.bgcolor is None:
            self.bgcolor = default.bgcolor
        if self.underline is None:
            self.underline = default.underline
        if self.bold is None:
            self.bold = default.bold
        if self.nowrap is None:
            self.nowrap = default.nowrap
        if self.rjust is None:
            self.rjust = default.rjust
        if self.fillrow is None:
            self.fillrow = default.fillrow


class Overlay(Style):

    def __init__(self, name, fgcolor=None, bgcolor=None,
                 underline=None, bold=None):
        super().__init__(name, fgcolor, bgcolor, underline, bold)

    def copy(self):
        return self.__class__(
            self.name,
            self.fgcolor,
            self.bgcolor,
            self.underline,
            self.bold
        )

    def set_default_attr(self, default):
        pass



from gappedbuf.re import *
from gappedbuf import re as gre


def match(pattern, doc, flags=0):
    """Try to apply the pattern at the start of the string, returning
    a match object, or None if no match was found."""
    return _compile(pattern, flags).match(doc)


def search(pattern, doc, flags=0):
    """Scan through string looking for a match to the pattern, returning
    a match object, or None if no match was found."""
    return _compile(pattern, flags).search(string)


def compile(pattern, flags=0):
    "Compile a regular expression pattern, returning a pattern object."
    return _compile(pattern, flags)


class _compile:

    def __init__(self, pattern, flags):
        self.reobj = gre._compile(pattern, flags)

    def findall(self, doc, *args, **kwargs):
        return self.reobj.findall(doc.buf, *args, **kwargs)

    def finditer(self, doc, *args, **kwargs):
        return self.reobj.finditer(doc.buf, *args, **kwargs)

    def match(self, doc, *args, **kwargs):
        return self.reobj.match(doc.buf, *args, **kwargs)

    def search(self, doc, *args, **kwargs):
        return self.reobj.search(doc.buf, *args, **kwargs)



import unicodedata
import kaa
from kaa.command import Commands
from kaa.command import commandid
from kaa.command import norec
from kaa.command import norerun
from kaa import document
from kaa.filetype.default import modebase
from kaa import doc_re


class CursorCommands(Commands):

    @commandid('cursor.right')
    @norerun
    def right(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        if wnd.screen.selection.has_mark():
            for i in range(wnd.get_command_repeat()):
                wnd.cursor.right()
        elif wnd.screen.selection.is_selected():
            selrange = wnd.screen.selection.get_selrange()
            if wnd.cursor.pos > selrange[1]:
                wnd.cursor.right()
            else:
                wnd.cursor.setpos(wnd.cursor.adjust_nextpos(
                    wnd.cursor.pos, selrange[1]))
        else:
            for i in range(wnd.get_command_repeat()):
                wnd.cursor.right()

        wnd.screen.selection.end_cursor_sel()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.right.select')
    @norerun
    def right_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor_sel(wnd.cursor.pos)
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.right()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.left')
    @norerun
    def left(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        if wnd.screen.selection.has_mark():
            for i in range(wnd.get_command_repeat()):
                wnd.cursor.left()
        elif wnd.screen.selection.is_selected():
            selrange = wnd.screen.selection.get_selrange()
            if wnd.cursor.pos < selrange[0]:
                wnd.cursor.left()
            else:
                wnd.cursor.setpos(wnd.cursor.adjust_nextpos(
                    wnd.cursor.pos, selrange[0]))
        else:
            for i in range(wnd.get_command_repeat()):
                wnd.cursor.left()

        wnd.screen.selection.end_cursor_sel()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.left.select')
    @norerun
    def left_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor_sel(wnd.cursor.pos)
        wnd.cursor.left()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.up')
    @norerun
    def up(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor_sel()
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.up()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.up.select')
    @norerun
    def up_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor_sel(wnd.cursor.pos)
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.up()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.down')
    @norerun
    def down(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor_sel()
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.down()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.down.select')
    @norerun
    def down_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor_sel(wnd.cursor.pos)
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.down()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.prev-line')
    @norerun
    def prev_line(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor_sel()
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.prev_line()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.next-line')
    @norerun
    def next_line(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor_sel()
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.next_line()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.word-right')
    @norerun
    def word_right(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor_sel()
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.right(word=True)
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.word-right.select')
    @norerun
    def word_right_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor_sel(wnd.cursor.pos)
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.right(word=True)
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.word-left')
    @norerun
    def word_left(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor_sel()
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.left(word=True)
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.word-left.select')
    @norerun
    def word_left_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor_sel(wnd.cursor.pos)
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.left(word=True)
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.pagedown')
    @norerun
    def pagedown(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor_sel()
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.pagedown()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.pagedown.select')
    @norerun
    def pagedown_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor_sel(wnd.cursor.pos)
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.pagedown()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.pageup')
    @norerun
    def pageup(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor_sel()
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.pageup()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.pageup.select')
    @norerun
    def pageup_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor_sel(wnd.cursor.pos)
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.pageup()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.home')
    @norerun
    def home(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor_sel()
        wnd.cursor.home()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.home.select')
    @norerun
    def home_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor_sel(wnd.cursor.pos)
        wnd.cursor.home()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.end')
    @norerun
    def end(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor_sel()
        wnd.cursor.end()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.end.select')
    @norerun
    def end_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor_sel(wnd.cursor.pos)
        wnd.cursor.end()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.top-of-line')
    @norerun
    def tol(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor_sel()
        wnd.cursor.tol(wnd.cursor.pos)
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.first-letter-of-line')
    @norerun
    def first_letter(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor_sel()
        wnd.cursor.first_letter(wnd.cursor.pos)
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.end-of-line')
    @norerun
    def eol(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor_sel()
        wnd.cursor.eol(wnd.cursor.pos)
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.top-of-file')
    @norerun
    def top(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor_sel()
        wnd.cursor.tof()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.top-of-file.select')
    @norerun
    def top_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor_sel(wnd.cursor.pos)
        wnd.cursor.tof()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.end-of-file')
    @norerun
    def last(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.end_cursor_sel()
        wnd.cursor.eof()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.end-of-file.select')
    @norerun
    def last_select(self, wnd):
        wnd.document.mode.cancel_auto_indent(wnd)
        wnd.screen.selection.begin_cursor_sel(wnd.cursor.pos)
        wnd.cursor.eof()
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('cursor.go-to-line')
    @norerun
    @norec
    def go_to_line(self, wnd):
        def callback(w, s):
            s = s.strip()
            try:
                lineno = int(s)
            except ValueError as e:
                kaa.app.messagebar.set_message(str(e))
                return

            if lineno == 0 or lineno > wnd.document.buf.lineno.linecount():
                kaa.app.messagebar.set_message('Enter valid line number.')
                return

            wnd.document.mode.cancel_auto_indent(wnd)
            wnd.screen.selection.end_cursor_sel()

            pos = wnd.document.get_lineno_pos(lineno)
            tol = wnd.document.gettol(pos)
            wnd.cursor.setpos(wnd.cursor.adjust_nextpos(wnd.cursor.pos, tol))

        from kaa.ui.inputline import inputlinemode
        doc = inputlinemode.InputlineMode.build('Line number:', callback,
                                                filter=inputlinemode.number_filter)
        kaa.app.messagebar.set_message("Enter line number")

        kaa.app.show_dialog(doc)


class SelectionCommands(Commands):

    @commandid('selection.begin-cursor')
    @norerun
    def selection_begin(self, wnd):
        wnd.screen.selection.begin_cursor_sel(wnd.cursor.pos)

    @commandid('selection.set-to')
    @norerun
    def selection_set_end(self, wnd):
        wnd.screen.selection.set_to(wnd.cursor.pos)

    @commandid('selection.end-cursor')
    @norerun
    def selection_clear(self, wnd):
        wnd.screen.selection.end_cursor_sel()

    @commandid('selection.set-mark')
    @norerun
    def selection_set_mark(self, wnd):
        if not wnd.screen.selection.has_mark():
            wnd.screen.selection.set_mark(wnd.cursor.pos)
        else:
            wnd.screen.selection.set_mark(None)

    @commandid('selection.set-linewise-mark')
    @norerun
    def selection_set_linewisemark(self, wnd):
        wnd.screen.selection.set_linewise_mark(wnd.cursor.pos)

    @commandid('selection.set-rectangle-mark')
    @norerun
    def selection_set_rectangle_mark(self, wnd):
        if not wnd.screen.selection.has_mark():
            wnd.screen.selection.set_rectangle_mark(wnd.cursor.pos)
        else:
            wnd.screen.selection.set_rectangle_mark(None)

    @commandid('selection.all')
    @norerun
    def select_all(self, wnd):
        f = wnd.cursor.adjust_nextpos(wnd.cursor.pos, 0)
        t = wnd.cursor.adjust_nextpos(wnd.cursor.pos, wnd.document.endpos())
        wnd.screen.selection.set_range(f, t)

    @commandid('selection.curline')
    @norerun
    def select_cur_line(self, wnd):
        tol = wnd.cursor.adjust_nextpos(
            wnd.cursor.pos,
            wnd.document.gettol(wnd.cursor.pos))
        eol = wnd.cursor.adjust_nextpos(
            wnd.cursor.pos,
            wnd.document.geteol(tol))

        wnd.screen.selection.set_range(tol, eol)

    @commandid('selection.curword')
    @norerun
    def select_cur_word(self, wnd):
        ret = wnd.document.mode.get_word_at(wnd.cursor.pos)
        if ret:
            f, t, cg = ret
            f = wnd.cursor.adjust_nextpos(wnd.cursor.pos, f)
            t = wnd.cursor.adjust_nextpos(wnd.cursor.pos, t)

            wnd.screen.selection.set_range(f, t)
            wnd.cursor.setpos(t)

    @commandid('selection.expand-sel')
    @norerun
    @norec
    def expand_sel(self, wnd):
        keys = wnd.editmode.last_command_keys
        L = len(keys)

        if L >= 3 and keys[-1] == keys[-2] == keys[-3]:
            self.select_all(wnd)
            if kaa.app.macro.recording:
                kaa.app.macro.record(1, self.select_all)

        elif L >= 2 and (keys[-1] == keys[-2]):
            self.select_cur_line(wnd)
            if kaa.app.macro.recording:
                kaa.app.macro.record(1, self.select_cur_line)

        else:
            self.select_cur_word(wnd)
            if kaa.app.macro.recording:
                kaa.app.macro.record(1, self.select_cur_word)

    @commandid('selection.line.begin')
    @norerun
    def lineselection_begin(self, wnd):
        tol = wnd.cursor.adjust_nextpos(
            wnd.cursor.pos, wnd.document.gettol(wnd.cursor.pos))
        wnd.screen.selection.begin_cursor_sel(tol)

    @commandid('selection.line.set-end')
    @norerun
    def lineselection_set_end(self, wnd):
        f = wnd.screen.get_start()
        if f is None:
            self.select_cur_line(wnd)
        else:
            pos = wnd.cursor.pos
            if pos < f:
                tol = wnd.cursor.adjust_nextpos(
                    wnd.document.gettol(pos))
                wnd.screen.selection.set_end(tol)
            else:
                eol = wnd.cursor.adjust_nextpos(
                    wnd.document.geteol(pos))
                wnd.screen.selection.set_end(eol)


class EditCommands(Commands):

    def delete_sel(self, wnd):
        if wnd.screen.selection.is_selected():
            if not wnd.screen.selection.is_rectangular():
                sel = wnd.screen.selection.get_selrange()
                wnd.screen.selection.clear()
                if sel:
                    f, t = sel
                    wnd.document.mode.delete_string(wnd, f, t)
                    return True
            else:
                return self._delete_rect_sel(wnd)

    def _delete_rect_sel(self, wnd):
        with wnd.document.undo_group():
            (posfrom, posto, colfrom, colto
             ) = wnd.screen.selection.get_rect_range()

            while posfrom < posto:
                sel = wnd.screen.selection.get_col_string(
                    posfrom, colfrom, colto)
                if sel:
                    f, t, org = sel
                    if org.endswith('\n'):
                        t = max(f, t - 1)
                    wnd.document.mode.delete_string(wnd, f, t)
                    posto -= (t - f)
                posfrom = wnd.document.geteol(posfrom)

            return True

    @commandid('edit.delete')
    def delete(self, wnd):
        if self.delete_sel(wnd):
            return

        pos = nextpos = wnd.cursor.pos
        for i in range(wnd.get_command_repeat()):
            nextpos = wnd.document.get_nextpos(nextpos)

        nextpos = wnd.cursor.adjust_nextpos(pos, nextpos)
        if pos < nextpos:
            wnd.document.mode.delete_string(wnd, pos, nextpos)

    @commandid('edit.delete.word')
    def delete_word(self, wnd):
        if self.delete_sel(wnd):
            return

        pos = wnd.cursor.pos
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.right(word=True)

        nextpos = wnd.cursor.pos
        if pos < nextpos:
            wnd.document.mode.delete_string(wnd, pos, nextpos)

    @commandid('edit.delete.line')
    def delete_line(self, wnd):
        pos = wnd.cursor.pos
        nextpos = wnd.cursor.adjust_nextpos(
            pos, wnd.document.get_line_to(pos))
        if pos < nextpos:
            wnd.document.mode.delete_string(wnd, pos, nextpos)

    @commandid('edit.delete.currentline')
    def delete_currentline(self, wnd):
        pos = wnd.cursor.pos
        f = t = wnd.cursor.adjust_nextpos(pos, wnd.document.gettol(pos))
        for i in range(wnd.get_command_repeat()):
            t = wnd.cursor.adjust_nextpos(pos, wnd.document.geteol(t))

        if f < t:
            wnd.document.mode.delete_string(wnd, f, t)

    @commandid('edit.backspace')
    def backspace(self, wnd):
        if self.delete_sel(wnd):
            return

        pos = prevpos = wnd.cursor.pos
        for i in range(wnd.get_command_repeat()):
            prevpos = wnd.cursor.adjust_nextpos(
                pos, wnd.document.get_prevpos(prevpos))

        if prevpos < pos:
            if pos == wnd.screen.pos:
                # locate cursor before delete to scroll half page up
                wnd.cursor.setpos(prevpos)

            wnd.document.mode.delete_string(wnd, prevpos, pos)

    @commandid('edit.backspace.word')
    def backspace_word(self, wnd):
        if self.delete_sel(wnd):
            return

        pos = wnd.cursor.pos
        for i in range(wnd.get_command_repeat()):
            wnd.cursor.left(word=True)

        prevpos = wnd.cursor.pos
        if prevpos < pos:
            if prevpos < wnd.screen.pos:
                # locate cursor before delete to scroll half page up
                wnd.cursor.setpos(prevpos)

            wnd.document.mode.delete_string(wnd, prevpos, pos)

    @commandid('edit.newline')
    @norerun
    def newline(self, wnd):
        if not wnd.document.mode.auto_indent:
            wnd.document.mode.put_string(wnd, '\n')
            wnd.screen.selection.clear()
            return

        wnd.screen.selection.clear()
        wnd.document.mode.on_auto_indent(wnd)

    def _indent_line(self, wnd, pos):
        mode = wnd.document.mode
        f, t = mode.get_indent_range(pos)
        if pos > t:
            wnd.document.mode.put_string(wnd, '\t')
            wnd.screen.selection.clear()
            return

        if f != t:
            cols = mode.calc_cols(f, t)
        else:
            cols = 0

        s = mode.build_indent_str(cols + mode.indent_width)
        mode.replace_string(wnd, f, t, s, True)

    @commandid('edit.indent')
    def indent(self, wnd):
        if not wnd.screen.selection.is_selected():
            self._indent_line(wnd, wnd.cursor.pos)
            return

        for i in range(wnd.get_command_repeat()):
            doc = wnd.document
            tol, eol = doc.mode.get_line_sel(wnd)
            wnd.screen.selection.set_range(tol, eol)

            with wnd.document.undo_group():
                mode = wnd.document.mode
                while tol < wnd.screen.selection.get_selrange()[1]:
                    f, t = mode.get_indent_range(tol)
                    if f != t:
                        cols = mode.calc_cols(f, t)
                    else:
                        cols = 0

                    endpos = doc.endpos()
                    eol = doc.geteol(tol)

                    if (t + 1 < eol != endpos) or (t < eol == endpos):
                        s = mode.build_indent_str(cols + mode.indent_width)
                        mode.replace_string(wnd, f, t, s, False)
                        tol = eol + (len(s) - (t - f))
                    else:
                        tol = eol

        f, t = wnd.screen.selection.get_selrange()
        wnd.cursor.setpos(f)
        wnd.cursor.savecol()

    def _dedent_line(self, wnd, pos):
        mode = wnd.document.mode
        f, t = mode.get_indent_range(pos)
        if f != t:
            cols = mode.calc_cols(f, t)
        else:
            cols = 0

        s = mode.build_indent_str(max(0, cols - mode.indent_width))
        mode.replace_string(wnd, f, t, s, True)

    @commandid('edit.dedent')
    def dedent(self, wnd):
        if not wnd.screen.selection.is_selected():
            self._dedent_line(wnd, wnd.cursor.pos)
            return

        for i in range(wnd.get_command_repeat()):
            doc = wnd.document
            tol, eol = doc.mode.get_line_sel(wnd)
            wnd.screen.selection.set_range(tol, eol)

            with wnd.document.undo_group():
                mode = wnd.document.mode
                while tol < wnd.screen.selection.get_selrange()[1]:
                    f, t = mode.get_indent_range(tol)
                    if f != t:
                        cols = mode.calc_cols(f, t)
                    else:
                        cols = 0

                    if cols:
                        s = mode.build_indent_str(
                            max(0, cols - mode.indent_width))
                        mode.replace_string(wnd, f, t, s, False)

                    tol = doc.geteol(tol)

        f, t = wnd.screen.selection.get_selrange()
        wnd.cursor.setpos(f)
        wnd.cursor.savecol()

    def _undo(self, wnd, rec):
        (action, args, kwargs) = rec
        if action == wnd.document.mode.UNDO_INSERT:
            pos, s, cur_pos, newpos = args
            wnd.document.delete(pos, pos + len(s))
            return cur_pos
        elif action == wnd.document.mode.UNDO_REPLACE:
            pos, posto, s, deled, cur_pos, newpos = args
            wnd.document.replace(pos, pos + len(s), deled)
            return cur_pos
        else:
            pos, posto, deled, cur_pos, newpos = args
            wnd.document.insert(pos, deled)

        return cur_pos

    @commandid('edit.undo')
    @norerun
    def undo(self, wnd):
        for i in range(wnd.get_command_repeat()):
            if wnd.document.undo and wnd.document.undo.can_undo():
                wnd.screen.selection.clear()
                pos = None
                for rec in wnd.document.undo.undo():
                    pos = self._undo(wnd, rec)

                if pos is not None:
                    wnd.cursor.setpos(pos)
                    wnd.cursor.savecol()

                wnd.document.mode.on_edited(wnd)

    def _redo(self, wnd, rec):
        (action, args, kwargs) = rec
        if action == wnd.document.mode.UNDO_INSERT:
            pos, s, cur_pos, newpos = args
            wnd.document.insert(pos, s)
            return newpos
        elif action == wnd.document.mode.UNDO_REPLACE:
            pos, posto, s, deled, cur_pos, newpos = args
            wnd.document.replace(pos, posto, s)
            return pos
        else:
            pos, posto, deled, cur_pos, newpos = args
            wnd.document.delete(pos, posto)

        return newpos

    @commandid('edit.redo')
    @norerun
    def redo(self, wnd):
        for i in range(wnd.get_command_repeat()):
            if wnd.document.undo and wnd.document.undo.can_redo():
                wnd.screen.selection.clear()
                pos = None
                for rec in wnd.document.undo.redo():
                    pos = self._redo(wnd, rec)

                if pos is not None:
                    wnd.cursor.setpos(pos)
                    wnd.cursor.savecol()

                wnd.document.mode.on_edited(wnd)

    def _get_sel(self, wnd):
        if wnd.screen.selection.is_selected():
            if not wnd.screen.selection.is_rectangular():
                f, t = wnd.screen.selection.get_selrange()
                return wnd.document.gettext(f, t)
            else:
                s = []
                (posfrom, posto, colfrom, colto
                 ) = wnd.screen.selection.get_rect_range()

                while posfrom < posto:
                    sel = wnd.screen.selection.get_col_string(
                        posfrom, colfrom, colto)
                    if sel:
                        f, t, org = sel
                        s.append(org.rstrip('\n'))
                    else:
                        s.append('')
                    posfrom = wnd.document.geteol(posfrom)

                return '\n'.join(s)

    def _copy_sel(self, wnd):
        sel = self._get_sel(wnd)
        if sel:
            kaa.app.clipboard.set(sel)
        return sel

    @commandid('edit.copy')
    def copy(self, wnd):
        self._copy_sel(wnd)
        wnd.screen.selection.set_mark(None)

    @commandid('edit.cut')
    def cut(self, wnd):
        if self._copy_sel(wnd):
            self.delete_sel(wnd)
        wnd.screen.selection.clear()

    @commandid('edit.paste')
    def paste(self, wnd):
        s = kaa.app.clipboard.get()
        if s:
            with wnd.document.undo_group():
                for i in range(wnd.get_command_repeat()):
                    wnd.document.mode.put_string(wnd, s)
                    wnd.screen.selection.clear()

    @commandid('edit.conv.upper')
    def conv_upper(self, wnd):
        s = self._get_sel(wnd)
        if s:
            wnd.document.mode.put_string(wnd, s.upper())
            wnd.screen.selection.clear()

    @commandid('edit.conv.lower')
    def conv_lower(self, wnd):
        s = self._get_sel(wnd)
        if s:
            wnd.document.mode.put_string(wnd, s.lower())
            wnd.screen.selection.clear()

    @commandid('edit.conv.nfkc')
    def conv_nfkc(self, wnd):
        s = self._get_sel(wnd)
        if s:
            wnd.document.mode.put_string(wnd, unicodedata.normalize('NFKC', s))
            wnd.screen.selection.clear()

    @commandid('edit.conv.full-width')
    def conv_fullwidth(self, wnd):
        import pyjf3
        s = self._get_sel(wnd)
        if s:
            wnd.document.mode.put_string(wnd, pyjf3.tofull(s))
            wnd.screen.selection.clear()

    @commandid('edit.word-complete')
    @norec
    @norerun
    def complete(self, wnd):
        from kaa.ui.wordcomplete import wordcompletemode
        wordcompletemode.show_wordlist(wnd)

    @commandid('edit.clipboard-history')
    @norec
    @norerun
    def clipboard_history(self, wnd):
        words = list(kaa.app.clipboard.get_all())
        if not words:
            return

        def callback(text):
            if text:
                wnd.document.mode.put_string(wnd, text)
                wnd.screen.selection.clear()
                kaa.app.clipboard.set(text)

        from kaa.ui.texthist import texthistmode
        texthistmode.show_history('Search clipboard text:', callback, words)


class CodeCommands(Commands):

    @commandid('code.region.linecomment')
    def linecomment(self, wnd):
        if not wnd.screen.selection.is_selected():
            tol = wnd.document.gettol(wnd.cursor.pos)
            wnd.document.mode.insert_string(
                wnd, tol, wnd.document.mode.LINE_COMMENT,
                update_cursor=False)
            wnd.cursor.setpos(tol)
            wnd.cursor.savecol()
            return
        else:
            tol, eol = wnd.document.mode.get_line_sel(wnd)
            wnd.screen.selection.set_range(tol, eol)

            with wnd.document.undo_group():
                while tol < wnd.screen.selection.get_selrange()[1]:
                    wnd.document.mode.insert_string(
                        wnd, tol, wnd.document.mode.LINE_COMMENT,
                        update_cursor=False)
                    tol = wnd.document.geteol(tol)

            f, t = wnd.screen.selection.get_selrange()
            wnd.cursor.setpos(f)
            wnd.cursor.savecol()

    def _is_comment_line(self, wnd, pos):
        reobj = doc_re.compile(r'[ \t]*({})'.format(
            doc_re.escape(wnd.document.mode.LINE_COMMENT)))
        return reobj.match(wnd.document, pos)

    @commandid('code.region.unlinecomment')
    def uncomment(self, wnd):
        if not wnd.screen.selection.is_selected():
            tol = wnd.document.gettol(wnd.cursor.pos)
            m = self._is_comment_line(wnd, tol)
            if m:
                f, t = m.span(1)
                wnd.document.mode.delete_string(
                    wnd, f, t)
            return
        else:
            tol, eol = wnd.document.mode.get_line_sel(wnd)
            wnd.screen.selection.set_range(tol, eol)

            with wnd.document.undo_group():
                while tol < wnd.screen.selection.get_selrange()[1]:
                    m = self._is_comment_line(wnd, tol)
                    if m:
                        f, t = m.span(1)
                        wnd.document.mode.delete_string(
                            wnd, f, t, update_cursor=False)
                    tol = wnd.document.geteol(tol)

            f, t = wnd.screen.selection.get_selrange()
            wnd.cursor.setpos(f)
            wnd.cursor.savecol()


class SearchCommands(Commands):

    @commandid('search.showsearch')
    @norec
    @norerun
    def showsearch(self, wnd):
        from kaa.ui.searchdlg import searchdlgmode

        doc = document.Document()
        mode = searchdlgmode.SearchDlgMode(target=wnd)
        doc.setmode(mode)
        mode.build_document()

        kaa.app.show_inputline(doc)

    @commandid('search.showreplace')
    @norec
    @norerun
    def showreplace(self, wnd):
        from kaa.ui.searchdlg import searchdlgmode

        doc = document.Document()
        mode = searchdlgmode.ReplaceDlgMode(target=wnd)
        doc.setmode(mode)
        mode.build_document()

        kaa.app.show_inputline(doc)

    def _show_searchresult(self, wnd, hit):
        if hit:
            f, t = hit.span()
            wnd.cursor.setpos(f)
            wnd.screen.selection.set_range(f, t)
            kaa.app.messagebar.set_message('found')
        else:
            kaa.app.messagebar.set_message('not found')

    @commandid('search.next')
    @norerun
    def searchnext(self, wnd):
        if not modebase.SearchOption.LAST_SEARCH:
            return
        if not modebase.SearchOption.LAST_SEARCH.text:
            return
        range = wnd.screen.selection.get_selrange()
        if range:
            start = range[0] + 1
        else:
            start = wnd.cursor.pos

        ret = wnd.document.mode.search_next(
            start, modebase.SearchOption.LAST_SEARCH)
        self._show_searchresult(wnd, ret)

        if not ret:
            if start != 0:
                ret = wnd.document.mode.search_next(
                    0, modebase.SearchOption.LAST_SEARCH)
                self._show_searchresult(wnd, ret)

    @commandid('search.prev')
    @norerun
    def searchprev(self, wnd):
        if not modebase.SearchOption.LAST_SEARCH:
            return
        if not modebase.SearchOption.LAST_SEARCH.text:
            return
        range = wnd.screen.selection.get_selrange()
        if range:
            start = range[1] - 1
        else:
            start = wnd.cursor.pos

        ret = wnd.document.mode.search_prev(
            start, modebase.SearchOption.LAST_SEARCH)
        self._show_searchresult(wnd, ret)

        if not ret:
            if start != wnd.document.endpos():
                ret = wnd.document.mode.search_prev(
                    wnd.document.endpos(),
                    modebase.SearchOption.LAST_SEARCH)
                self._show_searchresult(wnd, ret)



from unittest.mock import patch
import kaa_testutils
from kaa.filetype.rst import rstmode


class TestRstHighlight(kaa_testutils._TestDocBase):
    DEFAULTMODECLASS = rstmode.RstMode
    TOKENIZER = rstmode.RstMode.tokenizer

    def test_inline(self):

        doc = self._getdoc("**abc**")
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 7, 
            [self.TOKENIZER.tokens.strong] * 7)

        doc = self._getdoc("*abc*")
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 5, 
            [self.TOKENIZER.tokens.emphasis] * 5)

        doc = self._getdoc('``abc``')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 7, 
            [self.TOKENIZER.tokens.literal] * 7)

        doc = self._getdoc('`abc`_')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 6, 
            [self.TOKENIZER.tokens.interpreted] * 6)

        doc = self._getdoc('abc_')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 4, 
            [self.TOKENIZER.tokens.reference] * 4)

        doc = self._getdoc('_`abc`')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 6, 
            [self.TOKENIZER.tokens.target] * 6)

        doc = self._getdoc('|abc|')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 5, 
            [self.TOKENIZER.tokens.substitution] * 5)

        doc = self._getdoc('[abc]_')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 6, 
            [self.TOKENIZER.tokens.citation] * 6)

    def test_inline_delim(self):

        doc = self._getdoc('abc*abc')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 7, 
            [self.TOKENIZER.tokens.default] * 7)

        doc = self._getdoc('*abc*abc')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 7, 
            [self.TOKENIZER.tokens.emphasis] * 7)


    def test_header(self):
        doc = self._getdoc('--\nab\n---')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 9, 
            [self.TOKENIZER.tokens.header1] * 9)

    def test_block(self):
        doc = self._getdoc('abc:: \na')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 8, 
            [self.TOKENIZER.tokens.default] * 3 +
            [self.TOKENIZER.tokens.block] * 4 +
            [self.TOKENIZER.tokens.default])

    def test_directive(self):
        doc = self._getdoc('.. abc:: 111\n 222\n333')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 21, 
            [self.TOKENIZER.tokens.directive] * 18 +
            [self.TOKENIZER.tokens.default] * 3)


    def test_table(self):
        doc = self._getdoc('+=+\n| |')
        doc.mode.run_tokenizer(None)
        kaa_testutils.check_style(doc, 0, 7, 
            [self.TOKENIZER.tokens.table_border] * 4 +
            [self.TOKENIZER.tokens.table_row] * 3)




import kaa_testutils
from kaa.filetype.javascript.javascriptmode import JavaScriptMode


class TestJavascriptHighlight(kaa_testutils._TestDocBase):
    DEFAULTMODECLASS = JavaScriptMode

    def test_regex(self):
        doc = self._getdoc('/abc/')
        doc.mode.run_tokenizer(None)
        assert doc.styles.getints(0, 5) == [JavaScriptMode.tokenizer.tokens.regex.styleid_span]*5

    def test_not_regex(self):
        doc = self._getdoc('if a /abc/')
        doc.mode.run_tokenizer(None)

        assert doc.styles.getints(0, 2) == [JavaScriptMode.tokenizer.tokens.keyword.styleid_token]*2
        assert doc.styles.getints(2, 10) == [JavaScriptMode.tokenizer.styleid_default]*8



import os
import unicodedata
import sys
import kaa
import kaa.log
from kaa import document, encodingdef
from kaa.filetype.default import defaultmode
from kaa import consts


class FileStorage:

    def adjust_encoding(self, encoding):
        if encoding.lower() == 'shiftjis':
            return 'cp932'
        encoding = encodingdef.normalize_encname(encoding)
        return encoding

    def guess_japanese_encoding(self, filename):
        import pyjf3
        ret = pyjf3.guess(open(filename, 'rb').read())

        if ret == pyjf3.ASCII:
            return kaa.app.config.DEFAULT_ENCODING
        elif ret == pyjf3.UTF8:
            return 'utf-8'
        elif ret == pyjf3.SJIS:
            return 'cp932'
        elif ret == pyjf3.EUC:
            return 'euc-jp'
        elif ret == pyjf3.JIS:
            return 'iso-2022-jp'
        elif ret == pyjf3.UTF16_LE:
            return 'utf-16le'
        elif ret == pyjf3.UTF16_BE:
            return 'utf-16be'

    def get_textio(self, fileinfo, filemustexists=False):
        try:
            # use surrogateescape to preserve file contents intact.
            textio = open(fileinfo.fullpathname, 'r',
                          encoding=self.adjust_encoding(fileinfo.encoding),
                          errors='surrogateescape',
                          newline=fileinfo.nlchars)

        except FileNotFoundError:
            if filemustexists:
                raise
            return None

        return textio

    def get_fileinfo(self, filename, encoding=None, newline=None):
        if sys.platform == 'darwin':
            filename = unicodedata.normalize('NFC', filename)

        if encoding is None:
            encoding = kaa.app.config.DEFAULT_ENCODING

        if newline is None:
            newline = kaa.app.config.DEFAULT_NEWLINE

        fileinfo = FileInfo()

        fullpath = os.path.abspath(filename)
        dirname, filename = os.path.split(filename)

        fileinfo.storage = self
        fileinfo.fullpathname = fullpath
        fileinfo.dirname = dirname
        fileinfo.filename = filename

        try:
            fileinfo.stat = os.stat(fullpath)
        except FileNotFoundError:
            fileinfo.stat = None

        fileinfo.encoding = encoding
        fileinfo.newline = newline

        return fileinfo

    def listdir(self, dirname):
        dirs = []
        files = []
        for name in os.listdir(dirname):
            if os.path.isdir(os.path.join(dirname, name)):
                dirs.append(name)
            else:
                files.append(name)
        if sys.platform == 'darwin':
            dirs = [unicodedata.normalize('NFC', n) for n in dirs]
            files = [unicodedata.normalize('NFC', n) for n in files]
        return dirs, files

    def _save_hist(self, filename):
        abspath = os.path.abspath(filename)
        dirname = os.path.dirname(abspath)
        if dirname.endswith(os.pathsep):
            dirname = dirname[:-1]

        kaa.app.config.hist('filename').add(abspath)
        kaa.app.config.hist('dirname').add(dirname)
        kaa.app.last_dir = dirname

    def newfile(self, mode=None, s='', temporary=False):
        buf = document.Buffer()
        if s:
            buf.insert(0, s)

        doc = document.Document(buf)
        doc.temporary = temporary

        doc.title = '<Untitled {}>'.format(kaa.app.NUM_NEWFILE)
        kaa.app.NUM_NEWFILE += 1

        doc.fileinfo = FileInfo()
        if not mode:
            mode = defaultmode.DefaultMode()
        doc.setmode(mode)
        return doc

    def openfile(self, filename, encoding=None, newline=None, nohist=False,
                 filemustexists=False, modecls=None):

        fileinfo = self.get_fileinfo(filename, encoding, newline)
        modecls = modecls if modecls else select_mode(fileinfo.fullpathname)
        modecls.update_fileinfo(fileinfo)

        textio = self.get_textio(fileinfo, filemustexists)

        buf = document.Buffer()
        if textio:
            src = textio.read()
            if fileinfo.nlchars is not None:
                src = src.replace(fileinfo.nlchars, '\n')
            buf.insert(0, src)

            if not nohist:
                self._save_hist(filename)

        doc = document.Document(buf)
        doc.fileinfo = fileinfo

        doc.setmode(modecls())

        dir, file = os.path.split(fileinfo.fullpathname)
        if not dir.endswith(os.path.sep):
            dir += os.path.sep

        kaa.app.messagebar.set_message('Read from {}({})'.format(file, dir))

        kaa.app.config.hist_storage.flush()
        return doc

    def save_document(self, doc, filename, encoding=None, newline=None,
                      nohist=False):
        # todo: move most of codes to mode class.
        if not nohist:
            self._save_hist(filename)
        old_fileinfo = doc.fileinfo

        if not doc.fileinfo:
            fileinfo = kaa.app.storage.get_fileinfo(
                filename, encoding, newline)
            doc.fileinfo = fileinfo
        else:
            if encoding is not None:
                doc.fileinfo.encoding = encoding

            if doc.fileinfo.encoding is None:
                doc.fileinfo.encoding = kaa.app.config.DEFAULT_ENCODING

            if newline is not None:
                doc.fileinfo.newline = newline

            if doc.fileinfo.newline is None:
                doc.fileinfo.newline = kaa.app.config.DEFAULT_NEWLINE

        # update mode if file ext was
        old_ext = '' if not old_fileinfo else old_fileinfo.file_ext
        if old_ext.upper() != os.path.splitext(filename)[1].upper():
            modecls = select_mode(filename)
            if type(doc.mode) is not modecls:
                doc.setmode(modecls())

        doc.mode.update_fileinfo(doc.fileinfo, doc)

        with open(filename, 'w',
                  encoding=self.adjust_encoding(doc.fileinfo.encoding),
                  newline=consts.NEWLINE_CHARS[doc.fileinfo.newline],
                  errors='surrogateescape') as f:

            # TODO: save as another file and rename.
            # TODO: fsync
            f.write(doc.gettext(0, doc.endpos()))

        fileinfo = self.get_fileinfo(filename, doc.fileinfo.encoding,
                                     doc.fileinfo.newline)
        doc.mode.on_file_saved(fileinfo)

        doc.fileinfo = fileinfo
        doc.fileinfo.check_update()  # ensure timestamps updated

        doc.set_title(None)
        doc.temporary = False
        if doc.undo:
            doc.undo.saved()

        dir, file = os.path.split(fileinfo.fullpathname)
        if not dir.endswith(os.path.sep):
            dir += os.path.sep

        kaa.app.messagebar.set_message('Written to {}({})'.format(file, dir))

        kaa.app.config.hist_storage.flush()


class FileInfo:
    storage = None
    fullpathname = ''
    dirname = ''
    filename = ''
    stat = None
    encoding = None
    newline = None

    @property
    def nlchars(self):
        return consts.NEWLINE_CHARS[self.newline]

    @property
    def file_ext(self):
        return os.path.splitext(self.fullpathname)[1]

    def __init__(self):
        self.encoding = kaa.app.config.DEFAULT_ENCODING
        self.newline = kaa.app.config.DEFAULT_NEWLINE

    def save_screenloc(self, pos):
        if self.fullpathname:
            info = kaa.app.config.hist('filedisp').find(self.fullpathname)
            if not info:
                info = {}
            info['pos'] = pos
            kaa.app.config.hist('filedisp').add(
                self.fullpathname, info)

    def check_update(self):
        if self.fullpathname:
            try:
                stat = os.stat(self.fullpathname)
            except Exception:
                return

            if stat.st_mtime == self.stat.st_mtime:
                return

            if stat.st_size == self.stat.st_size:
                return

            self.stat = stat
            return True


def select_mode(filename):
    for pkg in kaa.app.config.get_mode_packages():
        if hasattr(pkg, 'FileTypeInfo'):
            ret = pkg.FileTypeInfo.select_mode(filename)
            if ret:
                return ret

    return defaultmode.DefaultMode



import unicodedata
import copy
from kaa.filetype.default import defaultmode
from kaa import doc_re
from kaa.theme import Style
from kaa.command import commandid
from kaa.command import norec
from kaa.command import norerun
from kaa.keyboard import *
from kaa.syntax_highlight import *

RstThemes = {
    'basic': [
        Style('header', 'Blue', None),
        Style('block', 'Orange', None),
        Style('directive', 'Green', None),
        Style('table', 'Cyan', None),
        Style('strong', 'Magenta', None),
        Style('emphasis', 'Blue', None),
        Style('literal', 'Cyan', None),
        Style('reference', 'Red', None),
        Style('role', 'Cyan', None),
        Style('substitution', 'Green', None),
    ],
}


def _build_seps():
    starts = []
    ends = []

    for i in range(0x12000):
        c = chr(i)
        category = unicodedata.category(c)
        if category in {'Pd', 'Po', 'Ps', 'Pi', 'Pf'}:
            starts.append(c)
        if category in {'Pd', 'Po', 'Pe', 'Pi', 'Pf'}:
            ends.append(c)
    return ''.join(starts), ''.join(ends)

SEP_STARTS, SEP_ENDS = _build_seps()


class RstInline(Span):
    WS = ' \t\r\n'
    STARTS = '\'"([{<' + WS + SEP_STARTS
    ENDS = '\'".,:;!?-)]}/\\>' + WS + SEP_ENDS

    def on_start(self, doc, match):
        pos = match.start()
        if ((pos and (doc.gettext(pos - 1, pos) not in self.STARTS)) or
                (pos >= doc.endpos() - 1) or
                (doc.gettext(pos + 1, pos + 2) in self.WS)):

            yield pos, pos + 1, self.tokenizer.styleid_default
            return pos + 1, self.terminates

        ret = yield from super().on_start(doc, match)
        return ret

    def _is_span_end(self, doc, m):
        pos = m.end()
        if pos < doc.endpos():
            return (doc.gettext(pos, pos + 1) in self.ENDS)
        return True


class TableToken(SingleToken):

    def on_start(self, doc, match):
        in_table = False
        pos = match.start()
        tol = doc.gettol(pos)
        if pos == tol:
            in_table = True
        else:
            # check if previous line is table
            if tol > 1:
                token = self.tokenizer.get_token_at(doc, tol - 2)
                if isinstance(token, TableToken):
                    in_table = True
        if in_table:
            ret = yield from super().on_start(doc, match)
            return ret
        else:
            yield pos, pos + 1, self.tokenizer.styleid_default
            return pos + 1, self.terminates


RST_HEADERS = r'=\-`:\'"~^_*+#<>'


def make_tokenizer():
    return Tokenizer(tokens=[
        # escape
        ('escape', SingleToken('default', [r'\\.'])),

        # header token
        ('header1', SingleToken('header',
                                [r'^(?P<H>[{}])(?P=H)+\n.+\n(?P=H)+$'.format(RST_HEADERS)])),
        ('header2', SingleToken('header',
                                [r'^.+\n(?P<H2>[{}])(?P=H2)+$'.format(RST_HEADERS)])),

        # block
        ('directive', Span('directive', r'\.\.\s+\S+::', '^\S',
                           escape='\\', capture_end=False)),
        ('block', Span('block', r'::', '^\S', capture_end=False)),

        # table
        ('table_border', TableToken('table', [r'\+[+\-=]+(\s+|$)'])),
        ('table_row', TableToken('table', [r'\|(\s+|$)'])),

        # inline token
        ('strong', RstInline('strong',
                             r'\*\*', r'\*\*', escape='\\')),
        ('emphasis', RstInline('emphasis',
                               r'\*', r'\*', escape='\\')),
        ('literal', RstInline('literal', r'``', r'``', escape='')),
        ('interpreted', RstInline('reference', r'`', r'`_?_?',
                                  escape='\\')),
        ('reference', SingleToken('reference',
                                  [r'\b[a-zA-Z0-9]+__?\b'])),
        ('role', RstInline('role', r':[a-zA-Z0-9]+:`', r'`',
                           escape='\\')),
        ('target', RstInline('reference',
                             r'_`\w', r'`', escape='\\')),
        ('substitution', SingleToken('substitution', [r'\|\w+\|'])),
        ('citation', SingleToken('reference', [r'\[\w+\]_\b'])),
    ])


RSTMENU = [
    ['&Table of contents', None, 'toc.showlist'],
]

rst_keys = {
    (ctrl, 't'): 'toc.showlist',
}


class RstMode(defaultmode.DefaultMode):
    MODENAME = 'Rst'
    tokenizer = make_tokenizer()

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, [rst_keys])

    def init_menu(self):
        super().init_menu()
        self.menu['CODE'] = copy.deepcopy(RSTMENU)

    def init_themes(self):
        super().init_themes()
        self.themes.append(RstThemes)

    HEADER1 = r'''^(?P<H>[{}])(?P=H)+\n
                  (?P<TITLE>.+)\n
                  (?P=H)+$'''.format(RST_HEADERS)

    HEADER2 = r'''^(?P<TITLE2>.+)\n
                (?P<H2>[{}])(?P=H2)+$'''.format(RST_HEADERS)

    RE_HEADER = doc_re.compile(
        '|'.join([HEADER1, HEADER2]), doc_re.X | doc_re.M)

    def get_headers(self):
        levels = {}
        stack = []
        pos = 0
        while True:
            m = self.RE_HEADER.search(self.document, pos)
            if not m:
                break

            pos = m.end()
            name = m.group('TITLE') or m.group('TITLE2')
            name = name.strip()

            bar = (m.group('H'), m.group('H2'))
            if bar not in levels:
                level = len(levels)
                levels[bar] = level
            else:
                level = levels[bar]

            if not stack:
                header = self.HeaderInfo('namespace', None, name, name,
                                         None, m.start())
                yield header
                stack.append((level, header))
                continue

            for i, (parent_level, info) in enumerate(stack):
                if level <= parent_level:
                    del stack[i:]
                    break

            if stack:
                parent = stack[-1][1]
            else:
                parent = None

            header = self.HeaderInfo(
                'namespace', parent, name,
                name, None, m.start())
            yield header
            stack.append((level, header))

    @commandid('toc.showlist')
    @norec
    @norerun
    def show_toclist(self, wnd):
        from kaa.ui.toclist import toclistmode
        headers = list(self.get_headers())
        toclistmode.show_toclist(wnd, headers)



import logging
import logging.handlers
import os


def init(logdir):
    logfile = os.path.join(logdir, 'kaa.log')

    logging.basicConfig(
        format="%(asctime)s %(name)s:%(levelname)s:%(message)s",
        datefmt='%Y/%m/%d %H:%M:%S',
        level=logging.DEBUG,
        handlers=[logging.handlers.RotatingFileHandler(
                filename=logfile, encoding='utf-8',
            maxBytes=10 * 1024 * 1024)])

    global LOG, debug, info, warning, error, critical, exception
    LOG = logging.getLogger('kaa')

    debug = LOG.debug
    info = LOG.info
    warning = LOG.warning
    error = LOG.error
    critical = LOG.critical
    exception = LOG.exception



import re
from kaa.filetype.default import defaultmode
from kaa import doc_re
from kaa.theme import Style
from kaa.filetype.javascript import javascriptmode
from kaa.filetype.css import cssmode

from kaa import encodingdef
from kaa.syntax_highlight import *


def iter_b_attr(b):
    pos = 0
    prop = re.compile(br'(\w+)\s*=\s*', re.DOTALL)
    value = re.compile(b'\s*(("[^"]+")|(\'[^\']+\'))', re.DOTALL)

    while True:
        m = prop.search(b, pos)
        if not m:
            return
        name = m.group(1)
        m = value.match(b, m.end())
        if not m:
            return
        yield name, m.group()[1:-1]
        pos = m.end()


def get_encoding(b):
    # http://www.w3.org/International/questions/qa-html-encoding-declarations

    # HTML5: <meta charset="UTF-8">
    m = re.search(br'<meta ', b, re.DOTALL)
    if m:
        for name, value in iter_b_attr(b[m.end():]):
            if name == b'charset':
                return str(value.strip(), 'utf-8')

    # HTML4, XHTML:
    #    <meta http-equiv="Content-type" content="text/html;charset=UTF-8">
    m = re.search(br'<meta ', b, re.DOTALL)
    if m:
        for name, value in iter_b_attr(b[m.end():]):
            if name == b'content':
                m = re.search(br'charset=(.*)', value, re.DOTALL)
                if m:
                    return str(m.group(1).strip(), 'utf-8')
                break

    # XHTML: <?xml version="1.0" encoding="UTF-8"?>
    m = re.search(br'<\?xml ', b, re.DOTALL)
    if m:
        for name, value in iter_b_attr(b[m.end():]):
            if name == b'encoding':
                return str(value.strip().strip(), 'utf-8')


HTMLThemes = {
    'basic': [
        Style('html-decl', 'Red', 'default'),
        Style('html-tag', 'Magenta', 'default'),
        Style('html-attrname', 'green', 'default'),
        Style('html-attrvalue', 'yellow', 'default'),
    ],
}


RE_ATTRNAME = doc_re.compile(
    r'>|(?P<ATTRNAME>[-._:a-zA-Z0-9]+)(?P<EQUAL>\s*=)?\s*')
RE_ATTRVALUE = doc_re.compile(r'\s*(?P<ATTRVALUE>({}))'.format(
    '|'.join(['[-._:a-zA-Z0-9]+', '(?P<Q1>"[^"]*")', "(?P<Q2>'[^']*')"])))


def iter_attrs(doc, pos):
    while True:
        m = RE_ATTRNAME.search(doc, pos)
        if not m:
            return pos
        if m.group() == '>':
            return m.start()

        attrname = m.group('ATTRNAME')
        pos = m.end()

        # has attribute value?
        if m.group('EQUAL'):
            # yield values after '='
            m = RE_ATTRVALUE.match(doc, pos)
            if m:
                pos = m.end()
                attrvalue = m.group('ATTRVALUE')
                if m.group('Q1') or m.group('Q2'):
                    attrvalue = attrvalue[1:-1]
                    yield attrname, attrvalue
                    continue

        yield attrname, None


class HTMLTag(SingleToken):

    def on_start(self, doc, match):
        pos, terminates = yield from super().on_start(doc, match)
        if pos < doc.endpos():
            pos = yield from self.tokenizer.AttrTokenizer.run(doc, pos)
            if pos < doc.endpos():
                c = doc.gettext(pos, pos + 1)
                if c == '>':
                    yield (pos, pos + 1, self.styleid_token)
                    pos += 1
        return pos, terminates


class HTMLScriptTag(HTMLTag):

    def on_start(self, doc, match):
        pos, terminates = yield from super().on_start(doc, match)

        for name, value in iter_attrs(doc, match.start()):
            if name.lower() == 'type':
                if value and value.lower() != 'text/javascript':
                    return pos, terminates

        pos = yield from self.tokenizer.JSTokenizer.run(doc, pos)
        return pos, terminates


class HTMLStyleTag(HTMLTag):

    def on_start(self, doc, match):
        pos, terminates = yield from super().on_start(doc, match)

        pos = yield from self.tokenizer.CSSTokenizer.run(doc, pos)

        return pos, terminates


class HTMLAttr(SingleToken):

    def __init__(self, stylename, value_stylename, tokens, terminates=False):
        super().__init__(stylename, tokens, terminates=terminates)
        self._value_stylename = value_stylename

    def prepare(self):
        super().prepare()
        self.register_styles([(self._value_stylename, "styleid_value")])

    def on_start(self, doc, match):
        pos, terminates = yield from super().on_start(doc, match)
        if not terminates and match.group(0)[-1].strip().endswith("="):
            if pos >= doc.endpos():
                return pos, terminates

            c = doc.gettext(pos, pos + 1)
            attrname = match.group('attrname')
            attrname = attrname.lower() if attrname else ''

            if c in {"'", '"'}:
                yield (pos, pos + 1, self.styleid_value)

            if attrname.startswith('on'):
                if c == "'":
                    pos = yield from self.tokenizer.AttrValueJSTokenizer1.run(doc, pos + 1)
                elif c == '"':
                    pos = yield from self.tokenizer.AttrValueJSTokenizer2.run(doc, pos + 1)

                if pos < doc.endpos():
                    if doc.gettext(pos, pos + 1) == c:
                        yield (pos, pos + 1, self.styleid_value)
                        pos += 1

            elif attrname == 'style':
                if c == "'":
                    pos = yield from self.tokenizer.AttrValueCSSTokenizer1.run(doc, pos + 1)
                elif c == '"':
                    pos = yield from self.tokenizer.AttrValueCSSTokenizer2.run(doc, pos + 1)

                if pos < doc.endpos():
                    if doc.gettext(pos, pos + 1) == c:
                        yield (pos, pos + 1, self.styleid_value)
                        pos += 1
            else:
                if c == "'":
                    pos = yield from self.tokenizer.AttrValueTokenizer1.run(doc, pos + 1)
                elif c == '"':
                    pos = yield from self.tokenizer.AttrValueTokenizer2.run(doc, pos + 1)
                else:
                    pos = yield from self.tokenizer.AttrValueTokenizer3.run(doc, pos)

        return pos, terminates


def make_tokenizer():
    ret = Tokenizer(tokens=[
        ('html-entityrefs', SingleToken('keyword',
                                        [r'&\w+;', r'&\#x[0-9a-hA-H]+;', r'&\#[0-9]+;'])),
        ('comment', Span('comment', r'<!--', r'--\s*>')),
        ('xmlpi', Span('html-decl', r'<\?', r'\?>')),
        ('xmldef', Span('html-decl', r'<!', r'>')),
        ('styletag', HTMLStyleTag('html-tag', [r'<\s*style'])),
        ('scripttag', HTMLScriptTag('html-tag', [r'<\s*script'])),
        ('closetag', SingleToken('html-tag', [r'</\s*[^>]+>'])),
        ('tag', HTMLTag('html-tag', [r'<\s*[^>\s]*'])),
    ])

    ret.AttrTokenizer = Tokenizer(parent=ret, terminates='>',
                                  is_resumable=False, tokens=[
                                      ('attr', HTMLAttr('html-attrname', 'html-attrvalue',
                                                        [r'(?P<attrname>[^\s=>]+)\s*=?\s*'])),
                                  ])

    ret.AttrTokenizer.AttrValueTokenizer1 = Tokenizer(
        parent=ret.AttrTokenizer,
        tokens=[('value', SingleToken('html-attrvalue', [r"[^']*'"], terminates=True))])

    ret.AttrTokenizer.AttrValueTokenizer2 = Tokenizer(
        parent=ret.AttrTokenizer,
        tokens=[('value', SingleToken('html-attrvalue', [r'[^"]*"'], terminates=True))])

    ret.AttrTokenizer.AttrValueTokenizer3 = Tokenizer(
        parent=ret.AttrTokenizer,
        tokens=[('value', SingleToken('html-attrvalue', [r'\w*'], terminates=True))])

    ret.AttrTokenizer.AttrValueJSTokenizer1 = Tokenizer(
        parent=ret.AttrTokenizer,
        tokens=javascriptmode.javascript_tokens(), terminates="'")

    ret.AttrTokenizer.AttrValueJSTokenizer2 = Tokenizer(
        parent=ret.AttrTokenizer,
        tokens=javascriptmode.javascript_tokens(), terminates='"')

    ret.AttrTokenizer.AttrValueCSSTokenizer1 = cssmode.make_prop_tokenizer(
        ret.AttrTokenizer, "'")
    ret.AttrTokenizer.AttrValueCSSTokenizer2 = cssmode.make_prop_tokenizer(
        ret.AttrTokenizer, '"')

    ret.JSTokenizer = Tokenizer(parent=ret,
                                tokens=javascriptmode.javascript_tokens(), terminates=r"</.*script[^>]*>")

    ret.CSSTokenizer = cssmode.make_tokenizer(
        ret, terminates=r"</.*style[^>]*>")

    return ret


class HTMLMode(defaultmode.DefaultMode):
    MODENAME = 'HTML'

    tokenizer = make_tokenizer()

    @classmethod
    def update_fileinfo(cls, fileinfo, document=None):
        try:
            buffer = open(fileinfo.fullpathname, 'rb').read(1024)
            enc = get_encoding(buffer)
            if enc:
                enc = encodingdef.normalize_encname(enc,
                                                    fileinfo.encoding)
                fileinfo.encoding = enc
        except IOError:
            pass

    def init_themes(self):
        super().init_themes()
        self.themes.append(HTMLThemes)
        self.themes.append(cssmode.CSSThemes)
        self.themes.append(javascriptmode.JavaScriptThemes)



class Cursor:

    def __init__(self, wnd):
        self.wnd = wnd
        self.pos = 0
        self.preferred_col = 0
        self.preferred_linecol = 0
        self.last_screenpos = (-1, -1)

    def refresh(self, top=None, middle=None, bottom=None,
                align_always=False):
        self.pos, y, x = self.wnd.locate_cursor(
            self.pos, top=top, middle=middle, bottom=bottom,
            align_always=align_always)
        assert self.pos is not None
        self.last_screenpos = (x, y)

    def setpos(self, pos, top=None, middle=None, bottom=None,
               align_always=False):
        assert pos is not None

        self.pos = pos
        self.refresh(top=top, middle=middle, bottom=bottom,
                     align_always=align_always)

    def savecol(self):
        """Update current preferred column"""

        y, x = self.wnd.screen.getrowcol(self.pos)
        self.preferred_col = x
        self.preferred_linecol = self.wnd.screen.get_cursorcol(self.pos)

    def up(self):
        # Ensure current position is displayed
        self.wnd.screen.locate(self.pos, middle=True, align_always=False)
        idx, row = self.wnd.screen.getrow(self.pos)
        if row.posfrom == 0:
            return

        if idx == self.wnd.screen.portfrom:
            # Scroll up a line
            if not self.wnd.lineup():
                # First line of text file
                return False
            idx, prevrow = self.wnd.screen.getrow(row.posfrom - 1)
        else:
            idx -= 1

        pos = self.wnd.screen.get_pos_above(idx, self.preferred_col)
        pos = self.adjust_nextpos(self.pos, pos)
        self.setpos(pos)

        return True

    def down(self):
        # Ensure current position is displayed
        self.wnd.screen.locate(self.pos, middle=True, align_always=False)
        idx, row = self.wnd.screen.getrow(self.pos)

        if self.wnd.screen.is_lastrow(row):
            return False

        if idx + 1 >= self.wnd.screen.portto:
            # Scroll down a line
            if not self.wnd.linedown():
                # last line of text file
                return False
            idx, nextrow = self.wnd.screen.getrow(row.posto)
        else:
            idx += 1

        pos = self.wnd.screen.get_pos_under(idx, self.preferred_col)
        pos = self.adjust_nextpos(self.pos, pos)
        self.setpos(pos)

        return True

    def prev_line(self):
        tol = self.wnd.document.gettol(self.pos)
        if tol != 0:
            self.wnd.screen.locate(tol - 1, top=True)

            prev = self.wnd.document.gettol(tol - 1)
            pos = self.wnd.screen.get_pos_at_cols(prev, self.preferred_linecol)
            pos = self.adjust_nextpos(self.pos, pos)
            self.wnd.screen.locate(pos, top=True)

            self.setpos(pos)
            return True

    def next_line(self):
        next = self.wnd.document.gettol(self.wnd.document.geteol(self.pos))
        if self.pos < next:
            self.wnd.screen.locate(next, bottom=True)

            pos = self.wnd.screen.get_pos_at_cols(next, self.preferred_linecol)
            pos = self.adjust_nextpos(self.pos, pos)
            self.wnd.screen.locate(pos, bottom=True)

            self.setpos(pos)
            return True

    def adjust_nextpos(self, curpos, nextpos):
        return nextpos

    def word_right_pos(self, pos):
        # Get next word break
        nextpos = self.wnd.document.endpos()
        for word in self.wnd.document.mode.split_word(pos):
            f, t, chars, cg = word
            nextpos = t
            if f == pos:  # first word
                # get next word
                continue
            if cg[0] == 'Z':  # skip white space
                continue
            nextpos = f
            break

        return self.adjust_nextpos(pos, nextpos)

    def word_left_pos(self, pos):
        # Get previous word break
        prevpos = tol = self.wnd.document.gettol(pos)
        if pos == prevpos:
            prevpos -= 1
        else:
            for f, t, chars, cg in self.wnd.document.mode.split_word(tol):
                #  This word is at after cursor pos
                if pos <= f:
                    break
                if cg[0] == 'Z':  # skip white space
                    continue
                prevpos = f

        return self.adjust_nextpos(pos, prevpos)

    def right(self, word=False):
        # Ensure current position is displayed
        self.wnd.screen.locate(self.pos, middle=True, align_always=False)

        if self.pos == self.wnd.document.endpos():
            return

        if not word:
            # Get right of current position
            nextpos = self.wnd.document.get_nextpos(self.pos)
            nextpos = self.adjust_nextpos(self.pos, nextpos)
        else:
            nextpos = self.word_right_pos(self.pos)

        if nextpos != self.pos:
            # Scroll down if next position is not visible
            while not self.wnd.screen.is_visible(nextpos):
                # scroll down
                if not self.wnd.linedown():
                    break

            self.setpos(nextpos)
            self.savecol()

    def left(self, word=False):
        # Ensure current position is displayed
        self.wnd.screen.locate(self.pos, middle=True, align_always=False)

        if self.pos == 0:
            return

        if not word:
            # Get left of current position
            prevpos = self.wnd.document.get_prevpos(self.pos)
            prevpos = self.adjust_nextpos(self.pos, prevpos)
        else:
            prevpos = self.word_left_pos(self.pos)

        if prevpos != self.pos:
            # Scroll up if next position is not visible
            while not self.wnd.screen.is_visible(prevpos):
                # scroll up
                if not self.wnd.lineup():
                    break

            self.setpos(prevpos)
            self.savecol()

    def pagedown(self):
        # Ensure current position is displayed
        self.wnd.screen.locate(self.pos, middle=True, align_always=False)

        idx, row = self.wnd.screen.getrow(self.pos)
        y = idx - self.wnd.screen.portfrom

        lastrow = self.wnd.screen.rows[self.wnd.screen.portto - 1]
        if lastrow.posto == self.wnd.document.endpos():
            nextpos = self.adjust_nextpos(self.pos, lastrow.posto)
            if self.pos != lastrow.posto:
                self.setpos(nextpos)
                return True

        if self.wnd.pagedown():
            idx = max(0, min(self.wnd.screen.portto - 1,
                             self.wnd.screen.portfrom + y))
            pos = self.wnd.screen.get_pos_under(idx, self.preferred_col)

            nextpos = self.adjust_nextpos(self.pos, pos)
            self.setpos(nextpos)

            return True
        else:
            pos = self.wnd.document.endpos()
            nextpos = self.adjust_nextpos(self.pos, pos)
            if nextpos != pos:
                self.setpos(nextpos)
                return True

    def pageup(self):
        # Ensure current position is displayed
        self.wnd.screen.locate(self.pos, middle=True, align_always=False)

        idx, row = self.wnd.screen.getrow(self.pos)
        y = idx - self.wnd.screen.portfrom

        if self.wnd.pageup():
            idx = max(0, min(self.wnd.screen.portto - 1,
                             self.wnd.screen.portfrom + y))
            pos = self.wnd.screen.get_pos_above(idx, self.preferred_col)
            nextpos = self.adjust_nextpos(self.pos, pos)
            self.setpos(nextpos)

            return True
        elif self.pos != 0:
            nextpos = self.adjust_nextpos(self.pos, 0)
            if nextpos != self.pos:
                self.setpos(nextpos)
                return True

    def home(self):
        # Ensure current position is displayed
        self.wnd.screen.locate(self.pos, middle=True, align_always=False)

        idx, row = self.wnd.screen.getrow(self.pos)
        nextpos = self.adjust_nextpos(self.pos, row.posfrom)
        self.setpos(nextpos)
        self.savecol()

    def end(self):
        # Ensure current position is displayed
        self.wnd.screen.locate(self.pos, middle=True, align_always=False)

        idx, row = self.wnd.screen.getrow(self.pos)
        if self.wnd.screen.is_lastrow(row):
            pos = self.wnd.document.endpos()
        else:
            pos = row.posto - 1

        nextpos = self.adjust_nextpos(self.pos, pos)
        self.setpos(nextpos)
        self.savecol()

    def tol(self, pos):
        tol = self.wnd.document.gettol(pos)
        nextpos = self.adjust_nextpos(self.pos, tol)
        self.wnd.screen.locate(nextpos, middle=True)
        self.setpos(nextpos)
        self.savecol()

    def first_letter(self, pos):
        f, tol = self.wnd.document.mode.get_indent_range(pos)
        nextpos = self.adjust_nextpos(self.pos, tol)
        self.wnd.screen.locate(nextpos, middle=True)
        self.setpos(nextpos)
        self.savecol()

    def eol(self, pos):
        eol = self.wnd.document.get_line_to(pos)
        nextpos = self.adjust_nextpos(self.pos, eol)
        self.wnd.screen.locate(nextpos, middle=True)
        self.setpos(nextpos)
        self.savecol()

    def tof(self):
        self.wnd.screen.locate(0, top=True)
        nextpos = self.adjust_nextpos(self.pos, 0)
        self.setpos(nextpos)
        self.savecol()

    def eof(self):
        nextpos = self.wnd.document.endpos()
        self.wnd.screen.locate(nextpos, middle=True)
        nextpos = self.adjust_nextpos(self.pos, nextpos)
        self.setpos(nextpos)
        self.savecol()



import re
from kaa.filetype.default import defaultmode
from kaa.syntax_highlight import *

JavaScriptThemes = {
    'basic': [],
}


KEYWORDS = ["break", "case", "catch", "continue", "debugger", "default",
            "delete", "do", "else", "finally", "for", "function", "if", "in",
            "instanceof", "new", "return", "switch", "this", "throw", "try",
            "typeof", "var", "void", "while", "with", "class", "enum", "export",
            "extends", "import", "super", "implements", "interface", "let",
            "package", "private", "protected", "public", "static", "yield", ]


class Regex(Span):
    RE_ENDOFTERM = re.compile(r'[a-zA-Z0-9.)]')

    def _is_regex(self, doc, pos):
        comments = (self.tokenizer.tokens.comment1,
                    self.tokenizer.tokens.comment2)
        not_terms = (self.tokenizer.tokens.keyword,)

        while pos > 0:
            pos -= 1
            token = self.tokenizer.get_token_at(doc, pos)

            if token.tokenizer is not self.tokenizer:
                break

            top = token.get_token_begin(doc, pos)

            # skip comment token
            if token in comments:
                pos = top
                continue

            # check if prev token is keywords
            if token in not_terms:
                break

            s = doc.gettext(top, pos + 1).strip()

            # skip white-space
            if not s:
                pos = top
                continue

            # check if last token is term or closing parenthesis
            m = self.RE_ENDOFTERM.match(s[-1])
            if not m:
                break

            # last token is term(literal, variable, expr, ...)
            return False

        return True

    def on_start(self, doc, match):
        pos = match.start()
        if self._is_regex(doc, pos):
            ret = yield from super().on_start(doc, match)
            return ret
        else:
            # This / is divide operator
            yield (pos, pos + 1, self.tokenizer.styleid_default)
            return pos + 1, False
        return ret


def javascript_tokens():
    return (
        ("comment1", Span('comment', r'/\*', '\*/', escape='\\')),
        ("comment2", Span('comment', r'//', '$', escape='\\')),

        ("keyword", Keywords('keyword', KEYWORDS)),
        ("number", SingleToken('number',
                               [r'\b[0-9]+(\.[0-9]*)*\b', r'\b\.[0-9]+\b'])),
        ("regex", Regex('string', r'/', r'/\w*', escape='\\')),


        ("string1", Span('string', '"', '"', escape='\\')),
        ("string2", Span('string', "'", "'", escape='\\')),
    )


def make_tokenizer():
    return Tokenizer(tokens=javascript_tokens())


class JavaScriptMode(defaultmode.DefaultMode):
    MODENAME = 'JavaScript'
    tokenizer = make_tokenizer()

    def init_themes(self):
        super().init_themes()
        self.themes.append(JavaScriptThemes)



import io
import copy
from kaa.filetype.default import defaultmode
from kaa.theme import Style
from kaa import doc_re
from kaa.command import commandid
from kaa.command import norec
from kaa.command import norerun
from kaa.keyboard import *
from kaa.ui.pythondebug import port
from kaa.syntax_highlight import *

PythonThemes = {
    'basic': [
        Style('python-bytes', 'blue', None),
    ],
}

PYTHONMENU = [
    ['&Comment', None, 'code.region.linecomment'],
    ['&Uncomment', None, 'code.region.unlinecomment'],
    ['&Table of contents', None, 'toc.showlist'],
    ['Toggle &Breakpoint', None, 'debugger.toggle.breakpoint'],
]

python_keys = {
    (ctrl, 't'): 'toc.showlist',
    f8: 'debugger.toggle.breakpoint',
}

KEYWORDS = ['and', 'as', 'assert', 'break', 'class', 'continue', 'def',
            'del', 'elif', 'else', 'except', 'finally', 'for', 'from',
            'global', 'if', 'import', 'in', 'is', 'lambda', 'nonlocal',
            'not', 'or', 'pass', 'raise', 'return', 'try', 'while',
            'with', 'yield']

CONSTANTS = ['False', 'None', 'True']


def make_tokenizer():
    return Tokenizer(tokens=(
        ("keyword", Keywords('keyword', KEYWORDS)),
        ("constant", Keywords('constant', CONSTANTS)),
        ("decorator", SingleToken('directive', [r'@\w[\w.]*'])),
        ("number", SingleToken('number',
                               [r'\b[0-9]+(\.[0-9]*)*\b', r'\b\.[0-9]+\b'])),
        ("comment", Span('comment', r'\#', '$', escape='\\')),

        ("string31", Span('string', '[rR]?"""', '"""', escape='\\')),
        ("string32", Span('string', "[rR]?'''", "'''", escape='\\')),
        ("string11", Span('string', '[rR]?"', '"', escape='\\')),
        ("string12", Span('string', "[rR]?'", "'", escape='\\')),

        ('bytes31', Span('python-bytes',
                         '([bB][rR]?|[rR]?[bB])"""', '"""', escape='\\')),
        ('bytes32', Span('python-bytes',
                         "([bB][rR]?|[rR]?[bB])'''", "'''", escape='\\')),
        ('bytes11', Span('python-bytes',
                         '([bB][rR]?|[rR]?[bB])"', '"', escape='\\')),
        ('bytes12', Span('python-bytes',
                         "([bB][rR]?|[rR]?[bB])'", "'", escape='\\')),
    ))


class PythonMode(defaultmode.DefaultMode):
    MODENAME = 'Python'
    re_begin_block = doc_re.compile(r"[^#]*:\s*(#.*)?$")
    LINE_COMMENT = '#'

    tokenizer = make_tokenizer()

    @classmethod
    def update_fileinfo(cls, fileinfo, document=None):
        import tokenize
        if not document:
            try:
                with open(fileinfo.fullpathname, 'rb') as buffer:
                    encoding, lines = tokenize.detect_encoding(buffer.readline)
                    fileinfo.encoding = encoding
            except IOError:
                pass
        else:
            s = document.gettext(0, 1024).encode('utf-8', errors='ignore')
            buffer = io.BytesIO(s)
            encoding, lines = tokenize.detect_encoding(buffer.readline)
            fileinfo.encoding = encoding

    def init_keybind(self):
        super().init_keybind()
        self.register_keys(self.keybind, [python_keys])

    def init_menu(self):
        super().init_menu()
        self.menu['CODE'] = copy.deepcopy(PYTHONMENU)

    def init_themes(self):
        super().init_themes()
        self.themes.append(PythonThemes)

    def on_set_document(self, document):
        super().on_set_document(document)
        self.display_breakpoint()

    def on_del_window(self, wnd):
        super().on_del_window(wnd)
        self.update_python_breakpoints()

    def get_line_overlays(self):
        ret = super().get_line_overlays()

        for k, v in self.get_breakpoints():
            ret[v] = 'breakpoint'
        return ret

    def on_file_saved(self, fileinfo):
        if self.document.get_filename() != fileinfo.fullpathname:
            self._clear_breakpoints()

        super().on_file_saved(fileinfo)

    RE_SEARCH_NAME = doc_re.compile(r'(\\.)|(\w+)|(\S)')

    def _py_getname(self, pos):
        while True:
            m = self.RE_SEARCH_NAME.search(self.document, pos)
            if not m:
                return self.document.endpos()
            if m.lastindex == 1:
                pos = m.end()
                continue
            elif m.lastindex == 2:
                return m.start(), m.end(), m.group()
            else:
                return m.start(), m.start(), ''

    def _py_def(self, m, token):
        pos, end, name = self._py_getname(m.end())
        if name:
            f, t = self.get_indent_range(m.start())
            cols = self.calc_cols(f, t)

            yield token, f, (cols, name)
        return end

    def _py_end_comment(self, m):
        close = r'(\\\n)|(\n)'
        reobj = doc_re.compile(close)

        pos = m.end()
        while True:
            m = reobj.search(self.document, pos)
            if not m:
                return self.document.endpos()
            if m.lastindex == 2:
                return m.end()
            else:
                pos = m.end()

    def _py_end_string(self, m):
        close = r'\\.|(?P<CLOSE>%s)' % m.group()
        reobj = doc_re.compile(close)

        pos = m.end()
        while True:
            m = reobj.search(self.document, pos)
            if not m:
                return self.document.endpos()
            if m.lastgroup == 'CLOSE':
                return m.end()
            else:
                pos = m.end()

    TOKENIZE_SEARCH_CLOSE = r'''
            (?P<BACKSLASH>\\.)|(?P<PARENTHESIS>[\{\[\(])|
            (?P<COMMENT>\#)|(?P<STRING>\"|\"\"\"|\'|\'\'\')'''

    def _py_close_parenthesis(self, m):
        close = '|(?P<CLOSE>\\%s)' % self.PARENSIS_PAIR[m.group()]
        reobj = doc_re.compile(self.TOKENIZE_SEARCH_CLOSE + close, doc_re.X)
        pos = m.end()
        while True:
            m = reobj.search(self.document, pos)
            if not m:
                return self.document.endpos()

            g = m.lastgroup
            if g == 'CLOSE':
                return m.end()
            elif g == 'BACKSLASH':
                pos = m.end()
            elif g == 'PARENTHESIS':
                pos = self._py_close_parenthesis(m)
            elif g == 'STRING':
                pos = self._py_end_string(m)
            elif g == 'COMMENT':
                pos = self._py_end_comment(m)
            else:
                assert 0

    RE_TOKENIZE_SEARCH = doc_re.compile(
        r'''(?P<BACKSLASH>\\.)|(?P<CLASS>\bclass\b)|
            (?P<DEF>\bdef\b)|(?P<PARENTHESIS>[\{\[\(])|
            (?P<COMMENT>\#)|(?P<STRING>\"\"\"|\"|\'\'\'|\')''', doc_re.X)

    def _parse_names(self, pos):
        while True:
            m = self.RE_TOKENIZE_SEARCH.search(self.document, pos)
            if not m:
                break
            g = m.lastgroup
            if g == 'BACKSLASH':
                pos = m.end()
            elif g == 'CLASS':
                pos = yield from self._py_def(m, 'class')
            elif g == 'DEF':
                pos = yield from self._py_def(m, 'def')
            elif g == 'PARENTHESIS':
                pos = self._py_close_parenthesis(m)
            elif g == 'STRING':
                pos = self._py_end_string(m)
            elif g == 'COMMENT':
                pos = self._py_end_comment(m)
            else:
                assert 0

    def _py_getfuncs(self):
        for token, pos, (indent, name) in self._parse_names(0):
            if token in {'class', 'def'}:
                yield token, pos, indent, name

    def get_headers(self):
        stack = []
        for token, pos, indent, name in self._py_getfuncs():
            dispname = name if token is 'class' else name + '()'
            headertype = 'namespace' if token == 'class' else 'function'
            if not stack:
                header = self.HeaderInfo(
                    headertype, None, name, dispname, None, pos)
                yield header
                stack.append((indent, header))
                continue

            for i, (parent_indent, info) in enumerate(stack):
                if indent <= parent_indent:
                    del stack[i:]
                    break

            if stack:
                parents = tuple(header for (_, header) in stack
                                if header.token == 'namespace')
                parent = parents[-1] if parents else None
                fullname = '.'.join([stack[-1][1].name, name])
                header = self.HeaderInfo(
                    headertype, parent, fullname,
                    dispname, None, pos)
            else:
                header = self.HeaderInfo(
                    headertype, None, name, dispname, None, pos)

            yield header
            stack.append((indent, header))

    @commandid('toc.showlist')
    @norec
    @norerun
    def show_toclist(self, wnd):
        from kaa.ui.toclist import toclistmode
        toclistmode.show_toclist(wnd, list(self.get_headers()))

    RE_SPLIT_LINE = doc_re.compile(
        r'''(?P<BACKSLASH>\\.)|(?P<OPEN_PARENTHESIS>[\{\[\(])|
            (?P<CLOSE_PARENTHESIS>[\}\]\)])|(?P<COMMENT>\#)|
            (?P<STRING>\"\"\"|\"|\'\'\'|\')|(?P<COLON>:)''',
        doc_re.X)

    def _get_indent_reasons(self, pos, posto):
        while pos < posto:
            m = self.RE_SPLIT_LINE.search(self.document, pos, posto)
            if not m:
                break

            g = m.lastgroup
            if g == 'BACKSLASH':
                pos = m.end()
            elif g == 'OPEN_PARENTHESIS':
                yield 'open_parenthesis', m.span(), m.group()
                pos = m.end()
            elif g == 'CLOSE_PARENTHESIS':
                yield 'close_parenthesis', m.span(), m.group()
                pos = m.end()
            elif g == 'COMMENT':
                pos = self._py_end_comment(m)
            elif g == 'STRING':
                pos = self._py_end_string(m)
            elif g == 'COLON':
                yield 'colon', m.span(), m.group()
                pos = m.end()
            else:
                assert 0

    def _calc_parenthesis_balance(self, tokens):
        p = 0
        for token, tokenpos, s in tokens:
            if token == 'open_parenthesis':
                p += 1
            elif token == 'close_parenthesis':
                p -= 1
        return p

    def _find_parenthesis_open_line(self, pos):
        while pos > 0:
            tol = self.document.gettol(pos)
            tokens = list(self._get_indent_reasons(tol, pos))
            p = self._calc_parenthesis_balance(tokens)
            if p > 0:
                return tol
            pos = tol - 1
        return None

    def calc_next_indent(self, pos):
        tol = self.document.gettol(pos)
        tokens = list(self._get_indent_reasons(tol, pos))
        if not tokens:
            return None

        f, t = self.get_indent_range(pos)
        t = min(t, pos)
        cols = self.calc_cols(f, t)

        if tokens[-1][0] == 'colon':
            return cols + self.indent_width

        p = self._calc_parenthesis_balance(tokens)
        if p > 0:
            return cols + self.indent_width
        elif p == 0:
            return None
        else:
            open_line = self._find_parenthesis_open_line(tol - 1)
            if open_line is None:
                return None
            f, t = self.get_indent_range(open_line)
            t = min(t, pos)
            return self.calc_cols(f, t)

    def get_breakpoints(self):
        for k, v in self.document.marks.items():
            if isinstance(k, port.BreakPoint):
                yield k, v

    def update_python_breakpoints(self):
        filename = self.document.get_filename()
        if not filename:
            return

        bps = []
        for k, v in self.get_breakpoints():
            lineno = self.document.buf.lineno.lineno(v)
            k.lineno = lineno
            bps.append(k)

        port.set_breakpoints(filename, bps)

    def _clear_breakpoints(self):
        bps = [k for k, v in self.get_breakpoints()]
        for bp in bps:
            del self.document.marks[bp]

    def display_breakpoint(self):
        self._clear_breakpoints()
        fname = self.document.get_filename()
        if fname:
            bps = port.get_breakpoints(fname)
            for bp in bps:
                pos = self.document.get_lineno_pos(bp.lineno)
                self.document.marks[bp] = pos

        self.document.style_updated()

    @commandid('debugger.toggle.breakpoint')
    @norec
    @norerun
    def toggle_breakpoint(self, wnd):
        pos = wnd.cursor.pos
        tol = self.document.gettol(pos)
        for bp, pos in self.get_breakpoints():
            if tol == self.document.gettol(pos):
                del self.document.marks[bp]
                break
        else:
            filename = self.document.get_filename()
            if filename:
                lineno = self.document.buf.lineno.lineno(tol)
                bp = port.BreakPoint(filename, lineno)
                self.document.marks[bp] = tol

        self.document.style_updated()



from kaa.keyboard import *

# Todo: Splitting key bind table does not make sense.
# Put them together.

# application commands
app_keys = {
    (alt, 'm'): 'app.mainmenu',
    (alt, '/'): 'app.mainmenu',
    f1: 'app.mainmenu',
    f9: 'app.global.prev',
    f10: 'app.global.next',
    (alt, 'w'): 'menu.window',
    (alt, 'z'): 'tools.suspend',
}

# default cursor commands
cursor_keys = {
    left: 'cursor.left',
    right: 'cursor.right',
    up: 'cursor.up',
    down: 'cursor.down',

    (ctrl, left): 'cursor.word-left',
    (ctrl, right): 'cursor.word-right',

    pagedown: 'cursor.pagedown',
    pageup: 'cursor.pageup',

    (shift, left): 'cursor.left.select',
    (shift, right): 'cursor.right.select',
    (shift, up): 'cursor.up.select',
    (shift, down): 'cursor.down.select',

    (shift, ctrl, left): 'cursor.word-left.select',
    (shift, ctrl, right): 'cursor.word-right.select',

    home: ('cursor.home'),
    end: ('cursor.end'),

    (shift, home): 'cursor.home.select',
    (shift, end): 'cursor.end.select',

    (ctrl, home): 'cursor.top-of-file',
    (ctrl, end): 'cursor.end-of-file',

    (shift, ctrl, home): 'cursor.top-of-file.select',
    (shift, ctrl, end): 'cursor.end-of-file.select',

    (alt, 'a'): 'selection.all',

    (alt, 'c'): 'selection.expand-sel',

    (ctrl, '@'): 'selection.set-mark',
    (alt, '#'): 'selection.set-rectangle-mark',
}


# edit commands
edit_command_keys = {
    backspace: 'edit.backspace',
    (ctrl, 'h'): 'edit.backspace',
    (alt, 'h'): 'edit.backspace.word',

    delete: 'edit.delete',
    (ctrl, 'd'): 'edit.delete',
    (alt, 'd'): 'edit.delete.word',

    (ctrl, 'k'): 'edit.delete.line',
    (alt, 'k'): 'edit.delete.currentline',

    '\r': 'edit.newline',
    '\n': 'edit.newline',

    (ctrl, 'x'): 'edit.cut',
    (ctrl, 'c'): 'edit.copy',
    (ctrl, 'v'): 'edit.paste',
    (ctrl, 'z'): 'edit.undo',
    (ctrl, 'y'): 'edit.redo',

    (alt, 'v'): 'edit.clipboard-history',
}

addtional_edit_command_keys = {
    (ctrl, 'g'): 'cursor.go-to-line',

    tab: 'edit.indent',
    (shift, tab): 'edit.dedent',

    ((ctrl, 'u'), (alt, '!')): 'tools.execute-shell-command',

    (ctrl, 'o'): 'edit.word-complete',
}

# macro commands
macro_command_keys = {
    f6: 'macro.toggle-record',
    f5: 'macro.run',
}

# rerun commands
rerun_keys = {
    (alt, '.'): 'command.rerun',
}


# search commands
search_command_keys = {
    (ctrl, 's'): 'search.showsearch',
    (alt, 's'): 'search.showreplace',
    f2: 'search.prev',
    f3: 'search.next',
}

# emacs like keys
emacs_keys = {
    (ctrl, 'b'): 'cursor.left',
    (ctrl, 'f'): 'cursor.right',
    (ctrl, 'p'): 'cursor.prev-line',
    (ctrl, 'n'): 'cursor.next-line',

    (alt, 'b'): 'cursor.word-left',
    (alt, 'f'): 'cursor.word-right',

    (ctrl, 'a'): 'cursor.top-of-line',
    (ctrl, 'e'): 'cursor.end-of-line',
    (ctrl, '^'): 'cursor.first-letter-of-line',

    (alt, 'n'): 'cursor.pagedown',
    (alt, 'p'): 'cursor.pageup',

    (alt, '<'): 'cursor.top-of-file',
    (alt, '>'): 'cursor.end-of-file',
}


# vi like commands
command_mode_keys = {
    # editmode change
    'i': 'editmode.insert',
    'R': 'editmode.replace',
    'A': ('editmode.insert', 'cursor.end-of-line'),
    'v': ('editmode.visual', 'selection.set-mark'),
    'V': ('editmode.visual-linewise', 'selection.set-linewise-mark'),

    # cursor command
    'h': 'cursor.left',
    'l': 'cursor.right',
    'k': 'cursor.up',
    'j': 'cursor.down',

    'w': 'cursor.word-right',
    'b': 'cursor.word-left',

    '0': 'cursor.top-of-line',
    '^': 'cursor.first-letter-of-line',
    '$': 'cursor.end-of-line',

    'gg': 'cursor.top-of-file',
    'G': 'cursor.end-of-file',

    (ctrl, 'b'): 'cursor.pageup',
    (ctrl, 'f'): 'cursor.pagedown',

    # edit
    'r': 'edit.replace-next-char',
    'x': 'edit.delete',
    'd': 'edit.delete-next-move',

    # undo/redo
    'u': 'edit.undo',
    (ctrl, 'r'): 'edit.redo',

    # clipboard
    'y': ('edit.copy', 'editmode.command'),
}

visual_mode_keys = {
    left: 'cursor.left',
    right: 'cursor.right',
    up: 'cursor.up',
    down: 'cursor.down',

    pagedown: 'cursor.pagedown',
    pageup: 'cursor.pageup',

    'h': 'cursor.left',
    'l': 'cursor.right',
    'k': 'cursor.up',
    'j': 'cursor.down',

    'gg': 'cursor.top-of-file',
    'G': 'cursor.end-of-file',

    'y': ('edit.copy', 'selection.end-cursor', 'editmode.command'),
}

visual_linewise_mode_keys = {
    up: 'cursor.prev-line',
    down: 'cursor.next-line',

    'k': 'cursor.prev-line',
    'j': 'cursor.next-line',

    'y': ('edit.copy', 'selection.end-cursor', 'editmode.command'),
}




MENUS = {
    'MAIN':
    [['[&File]', 'FILE', None],
     ['[&Edit]', 'EDIT', None],
     ['[&Code]', 'CODE', None],
     ['[&Macro]', 'MACRO', None],
     ['[&Tools]', 'TOOLS', None],
     ['[&Git]', 'GIT', None],
     ['[&Window]', 'WINDOW', None]],

    'FILE':
    [['&New', None, 'file.new'],
     ['&Open', None, 'file.open'],
     ['File &Info', None, 'file.info'],
     ['View &Diff', None, 'file.viewdiff'],
     ['&Save', None, 'file.save'],
     ['Save &As', None, 'file.saveas'],
     ['&Close', None, 'file.close'],
     ['Save a&ll', None, 'file.save.all'],
     ['Clos&e all', None, 'file.close.all'],
     ['[&Recently]', 'RECENTLY-USED-FILES', None],
     ['&Quit', None, 'file.quit']],

    'RECENTLY-USED-FILES':
    [['Recently used &Files', None, 'file.recently-used-files'],
     ['Recently used &Dirs', None, 'file.recently-used-directories']],

    'EDIT':
    [['&Cut', None, 'edit.cut'],
     ['C&opy', None, 'edit.copy'],
     ['&Paste', None, 'edit.paste'],
     ['Paste &History', None, 'edit.clipboard-history'],
     ['&Undo', None, 'edit.undo'],
     ['&Redo', None, 'edit.redo'],
     ['&Search', None, 'search.showsearch'],
     ['R&eplace', None, 'search.showreplace'],
     ['Co&mplete', None, 'edit.word-complete'],
     ['[&Convert]', 'EDIT-CONVERT', None]],

    'EDIT-CONVERT':
    [['&Upper', None, 'edit.conv.upper'],
     ['&Lower', None, 'edit.conv.lower'],
     ['&Normalization', None, 'edit.conv.nfkc'],
     ['&Full-width', None, 'edit.conv.full-width']],

    'MACRO':
    [['&Start Record', None, 'macro.start-record'],
     ['&End Record', None, 'macro.end-record'],
     ['&Run Macro', None, 'macro.run']],

    'TOOLS':
    [['&Grep', None, 'tools.grep'],
     ['Paste &Lines', None, 'edit.paste-lines'],
     ['&Shell command', None, 'tools.execute-shell-command'],
     ['&Make', None, 'tools.make'],
     ['Spell &Checker', None, 'tools.spellchecker'],
     ['&Python console', None, 'python.console'],
     ['Python &Debugger', None, 'python.debugger.run'],
     ['Python debugger s&erver', None, 'python.debugger.server'],
     ],

    'GIT':
    [['&Status', None, 'git.status'],
     ['&Log', None, 'git.log'],
     ],

    'WINDOW':
    [['&Frame list', None, 'app.show-framelist'],
     ['Split &vert', None, 'editor.splitvert'],
     ['Split &horz', None, 'editor.splithorz'],
     ['&Move separator', None, 'editor.moveseparator'],
     ['&Next window', None, 'editor.nextwindow'],
     ['&Prev window', None, 'editor.prevwindow'],
     ['&Join window', None, 'editor.joinwindow'],
     ['[&Switch file]', 'CHANGE-WINDOW', None]],

    'CHANGE-WINDOW':
    [['&Switch file', None, 'editor.switchfile'],
     ['&New file here', None, 'file.new-to'],
     ['&Open file here', None, 'file.open-to'],
     ['Recently used &Files', None, 'file.recently-used-files-to'],
     ['Recently used &Dirs', None, 'file.recently-used-directories-to']],
}
